import pytest
from sdmx.model.common import Code, Codelist

from genno import Computer
from genno.compat.sdmx import codelist_to_groups
from genno.testing import add_test_data


def test_codelist_to_groups() -> None:
    c = Computer()
    _, t_foo, t_bar, __ = add_test_data(c)

    cl: Codelist = Codelist(id="t")
    cl.append(Code(id="foo", child=[Code(id=t) for t in t_foo]))
    cl.append(Code(id="bar", child=[Code(id=t) for t in t_bar]))

    # Operator runs
    for result0 in (
        codelist_to_groups(cl),
        codelist_to_groups(iter(cl), dim="t"),
    ):
        # Result has the expected contents
        assert {"t"} == set(result0.keys())
        result_t = result0["t"]
        assert {"foo", "bar"} == set(result_t.keys())
        assert set(t_foo) == set(result_t["foo"])
        assert set(t_bar) == set(result_t["bar"])

    with pytest.raises(ValueError, match="Must provide a dimension"):
        codelist_to_groups(iter(cl))

    # Output is usable in Computer() with aggregate
    c.require_compat("genno.compat.sdmx")
    c.add("t::codes", cl)
    c.add("t::groups", "codelist_to_groups", "t::codes")
    key = c.add("x::agg", "aggregate", "x:t-y", "t::groups", False)

    result1 = c.get(key)

    # Quantity was aggregated per `cl`
    assert {"foo", "bar"} == set(result1.coords["t"].data)



from dask.core import quote

from .config import configure
from .core.computer import Computer
from .core.exceptions import ComputationError, KeyExistsError, MissingKeyError
from .core.key import Key
from .core.operator import Operator
from .core.quantity import Quantity

__all__ = [
    "ComputationError",
    "Computer",
    "Key",
    "KeyExistsError",
    "MissingKeyError",
    "Operator",
    "Quantity",
    "configure",
    "quote",
]



import logging
from functools import partial
from inspect import Parameter, signature
from typing import Callable, Iterable, Mapping, MutableMapping, Tuple, Type, Union

import numpy as np
import pandas as pd
import pint
from dask.core import literal

from .compat.pint import PintError
from .compat.xarray import is_scalar
from .core.key import Key

log = logging.getLogger(__name__)


#: Replacements to apply to Quantity units before parsing by
#: :doc:`pint <pint:index>`. Mapping from original unit -> preferred unit.
#:
#: The default values include:
#:
#: - The '%' symbol cannot be supported by pint, because it is a Python operator; it is
#:   replaced with “percent”.
#:
#: Additional values can be added with :func:`configure`; see :ref:`config-units`.
REPLACE_UNITS = {
    "%": "percent",
}

# For use in type hints
UnitLike = Union[str, pint.Unit, pint.Quantity]


def clean_units(input_string):
    """Tolerate messy strings for units.

    - Dimensions enclosed in “[]” have these characters stripped.
    - Replacements from :data:`.REPLACE_UNITS` are applied.
    """
    input_string = input_string.strip("[]")
    for old, new in REPLACE_UNITS.items():
        input_string = input_string.replace(old, new)
    return input_string


def collect_units(*args):
    """Return the "_unit" attributes of the `args`."""
    registry = pint.get_application_registry()

    result = []
    for arg in args:
        try:
            unit = arg.attrs.get("_unit")
        except AttributeError:
            if is_scalar(arg):
                result.append(registry.dimensionless)
            else:
                raise  # pragma: no cover
        else:
            if unit is None:
                log.debug(
                    f"{arg.__class__.__name__} '{arg.name or '(no name)'}' {arg.dims!r}"
                    " lacks units; assume dimensionless"
                )
                unit = registry.dimensionless

            # Convert a possible string or other expression to a pint.Unit object
            arg.units = registry.Unit(unit)

            result.append(arg.units)

    return tuple(result)


def filter_concat_args(args):
    """Filter out str and Key from *args*.

    A warning is logged for each element removed.
    """
    for arg in args:
        if isinstance(arg, (str, Key)):
            log.warning(f"concat() argument {repr(arg)} missing; will be omitted")
            continue
        yield arg


def _invalid(unit: str, exc: Exception) -> Exception:
    """Helper method to return an intelligible exception from :func:`parse_units`."""
    chars = "".join(filter("-?$".__contains__, unit))
    msg = f"unit {unit!r} cannot be parsed; contains invalid character(s) {chars!r}"
    # Use the original class of `exc`, mapped in some cases
    cls_map: Mapping[Type[Exception], Type[Exception]] = {TypeError: ValueError}
    return_cls = cls_map.get(type(exc), type(exc))
    return return_cls(msg)


def parse_units(data: Iterable, registry=None) -> pint.Unit:
    """Return a :class:`pint.Unit` for an iterable of strings.

    Valid unit expressions not already present in the `registry` are defined, e.g.:

    .. code-block:: python

       u = parse_units(["foo/bar", "foo/bar"], reg)

    …results in the addition of unit definitions equivalent to:

    .. code-block:: python

       reg.define("foo = [foo]")
       reg.define("bar = [bar]")
       u = reg.foo / reg.bar

    Raises
    ------
    ValueError
        if `data` contains more than 1 unit expression, or the unit expression contains
        characters not parseable by :mod:`pint`, e.g. ``-?$``.
    """
    registry = registry or pint.get_application_registry()

    # Ensure a type that is accepted by pd.unique()
    if isinstance(data, str):
        data = np.array([data])
    elif not isinstance(data, (np.ndarray, pd.Index, pd.Series)):
        data = np.array(data)

    unit = pd.unique(data)

    if len(unit) > 1:
        raise ValueError(f"mixed units {list(unit)}")

    try:
        unit = clean_units(unit[0])
    except IndexError:
        # `units_series` is length 0 → no data → dimensionless
        unit = registry.dimensionless

    # Parse units
    try:
        return registry.Unit(unit)
    except pint.UndefinedUnitError:
        try:
            # Unit(s) do not exist; define them in the UnitRegistry
            # TODO add global configuration to disable this feature.
            # Split possible compound units
            for part in unit.split("/"):
                try:
                    registry.Unit(part)
                except pint.UndefinedUnitError:
                    # Part was unparseable; define it
                    definition = f"{part} = [{part}]"
                    log.info(f"Add unit definition: {definition}")

                    # This line will fail silently for parts like 'G$' containing
                    # characters like '$' that are discarded by pint
                    registry.define(definition)

            # Try to parse again
            return registry.Unit(unit)
        except PintError as e:
            # registry.define() failed somehow
            raise _invalid(unit, e)
    except (AttributeError, TypeError) + PintError as e:  # type: ignore [misc]
        # Unit contains a character like '-' that throws off pint
        # NB this 'except' clause must be *after* UndefinedUnitError, since that is a
        #    subclass of AttributeError.
        raise _invalid(unit, e)


def partial_split(func: Callable, kwargs: Mapping) -> Tuple[Callable, MutableMapping]:
    """Forgiving version of :func:`functools.partial`.

    Returns a :class:`partial` object and leftover kwargs not applicable to `func`.
    """
    # Names of parameters to `func`
    try:
        par_names: Mapping = signature(func).parameters
    except ValueError:
        # signature() raises for operator.itemgetter(…), built-ins, and similar
        if not callable(func):  # pragma: no cover
            raise TypeError(type(func))
        par_names = {}

    func_args, extra = {}, {}
    for name, value in kwargs.items():
        if name in par_names and par_names[name].kind in (
            Parameter.POSITIONAL_OR_KEYWORD,
            Parameter.KEYWORD_ONLY,
        ):
            # A keyword argument of `func`
            func_args[name] = value
        else:
            extra[name] = value

    if func_args:
        return partial(func, **func_args), extra
    else:
        return func, extra  # Nothing to partial; return `func` as-is


def unquote(value):
    """Reverse :func:`dask.core.quote`."""
    if isinstance(value, tuple) and len(value) == 1 and isinstance(value[0], literal):
        return value[0].data
    else:
        return value



import logging
from copy import copy
from functools import partial
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    MutableMapping,
    Optional,
    Tuple,
    Type,
    Union,
)

import pint
import yaml

import genno.computations as computations
from genno.core.computer import Computer
from genno.core.exceptions import KeyExistsError, MissingKeyError
from genno.core.key import Key, iter_keys
from genno.util import REPLACE_UNITS

log = logging.getLogger(__name__)

#: Registry of configuration section handlers.
HANDLERS: Dict[str, Any] = {}

#: Configuration sections/keys to be stored with no action.
STORE = set(["cache_path", "cache_skip"])


def configure(path: Optional[Union[Path, str]] = None, **config):
    """Configure :mod:`.genno` globally.

    Modifies global variables that affect the behaviour of *all* Computers and
    operators. Configuration keys loaded from file are superseded by keyword arguments.
    Messages are logged at level :obj:`logging.INFO` if `config` contains unhandled
    sections.

    Parameters
    ----------
    path : pathlib.Path, optional
        Path to a configuration file in JSON or YAML format.
    **config :
        Configuration keys/sections and values.
    """
    if path:
        config["path"] = Path(path)
    # The value of fail= doesn't matter, since no Computer is given
    parse_config(None, data=config, fail="raise")


def handles(section_name: str, iterate: bool = True, discard: bool = True):
    """Decorator to register a configuration section handler in :data:`HANDLERS`.

    Parameters
    ----------
    section_name: str
        The name of the configuration section to handle. Using a name already present
        in :data:`HANDLERS` overrides that handler.
    iterate : bool, optional
        If :obj:`True`, the handler is called once for each item (either list item, or
        (key, value) tuple) in the section. If :obj:`False`, the entire section
        contents, of whatever type, are passed to tha handler.
    discard : bool, optional
        If :obj:`True`, configuration section data is discarded after the handler is
        called. If :obj:`False`, the data is retained and stored on the Computer.
    """

    def wrapper(f: Callable):
        if section_name in HANDLERS:
            log.debug(
                f"Override handler {repr(HANDLERS[section_name])} for "
                f" '{section_name}:'"
            )
        HANDLERS[section_name] = f
        setattr(f, "_iterate", iterate)
        setattr(f, "_discard", discard)
        return f

    return wrapper


def parse_config(
    c: Optional[Computer],
    data: MutableMapping[str, Any],
    fail: Optional[Union[str, int]] = None,
):
    # Assemble a queue of (args, kwargs) for Computer.add_queue()
    queue: List[Tuple[Tuple, Dict]] = []

    try:
        path = data.pop("path")
    except KeyError:
        pass
    else:
        # Load configuration from file
        path = Path(path)
        with open(path, "r") as f:
            new_data = yaml.safe_load(f)

        # Overwrite the file content with direct configuration values
        new_data.update(data)
        data = new_data

        # Also store the directory where the configuration file was located
        if c is None:
            data["config_dir"] = path.parent
        else:
            # Early add to the graph
            c.graph["config"]["config_dir"] = path.parent

    # Sections to discard, e.g. with handler._store = False
    discard = set()

    for section_name, section_data in data.items():
        handler = HANDLERS.get(section_name)
        if not handler:
            if section_name not in STORE:
                log.info(
                    f"No handler for configuration section '{section_name}:'; ignored"
                )
            continue

        if handler._discard:
            discard.add(section_name)

        if handler._iterate:
            if isinstance(section_data, dict):
                iterator: Iterable = section_data.items()
            elif isinstance(section_data, list):
                iterator = section_data
            else:  # pragma: no cover
                raise NotImplementedError(handler.expected_type)
            queue.extend((("apply", handler), dict(info=item)) for item in iterator)
        else:
            handler(c, section_data)

    for section_name in discard:
        data.pop(section_name)

    if c:
        # Process the entries
        c.add_queue(queue, max_tries=2, fail=fail)

        # Store configuration in the graph itself
        c.graph["config"].update(data)
    elif len(queue):
        raise RuntimeError("Cannot apply non-global configuration without a Computer")


@handles("aggregate")
def aggregate(c: Computer, info):
    """Handle one entry from the ``aggregate:`` config section."""
    # Copy for destructive .pop()
    info = copy(info)

    # Unpack `info`
    quantities = c.infer_keys(info.pop("_quantities"))
    tag = info.pop("_tag")
    # Keyword arguments for add()
    kw = dict(
        fail=info.pop("_fail", None),
        groups={info.pop("_dim"): info},
        strict=True,
        sums=True,
    )

    def _log_or_raise(exc: Exception, default_level: str, message: str):
        """Either raise `exc` if ``kw["fail"]`` > `default_level`, or log `message`."""
        fail_level = getattr(logging, (kw["fail"] or default_level).upper())
        if fail_level >= logging.ERROR:
            raise exc
        else:
            log.log(fail_level, message)

    try:
        quantities = c.check_keys(*quantities)
    except MissingKeyError as e:
        # Default to fail="error" here: stricter
        _log_or_raise(e, "error", f"No key(s) {e.args!r} to aggregate")

    # Iterate over quantities to be aggregated
    for qty in map(Key, quantities):
        try:
            result = c.add(qty.add_tag(tag), "aggregate", qty, **kw)
        except KeyExistsError:
            pass
        except MissingKeyError as e:
            # Default to fail="warning": more permissive
            _log_or_raise(e, "warning", repr(e))
        else:
            if keys := list(iter_keys(result)):
                log.info(f"Add {repr(keys[0])} + {len(keys)-1} partial sums")


@handles("alias")
def alias(c: Computer, info):
    """Handle one entry from the ``alias:`` config section."""
    c.add(info[0], info[1])


@handles("combine")
def combine(c: Computer, info):
    """Handle one entry from the ``combine:`` config section."""
    # Split inputs into three lists
    quantities, select, weights = [], [], []

    # Key for the new quantity
    key = Key(info["key"])

    # Loop over inputs to the combination
    for i in info["inputs"]:
        # Required dimensions for this input: output key's dims, plus any
        # dims that must be selected on
        selector = i.get("select", {})
        dims = set(key.dims) | set(selector.keys())
        quantities.append(c.infer_keys(i["quantity"], dims))

        select.append(selector)
        weights.append(i.get("weight", 1))

    # Check for malformed input
    assert len(quantities) == len(select) == len(weights)

    # Computation
    task = tuple(
        [partial(computations.combine, select=select, weights=weights)] + quantities
    )

    added = iter_keys(c.add(key, task, strict=True, sums=True))

    log.info(f"Add {repr(key)} + {len(list(added))-1} partial sums")
    log.debug("    as combination of")
    log.debug(f"    {repr(quantities)}")


@handles("default", iterate=False)
def default(c: Computer, info):
    """Handle the ``default:`` config section."""
    c.default_key = info


@handles("files")
def files(c: Computer, info):
    """Handle one entry from the ``files:`` config section."""
    # Files with exogenous data
    path = Path(info["path"])
    if not path.is_absolute():
        # Resolve relative paths relative to the directory containing the configuration
        # file
        path = c.graph["config"].get("config_dir", Path.cwd()) / path

    info["path"] = path

    c.add("load_file", **info)


@handles("general")
def general(c: Computer, info):
    """Handle one entry from the ``general:`` config section."""
    # Inputs
    inputs = c.infer_keys(info.get("inputs", []))

    if info["comp"] in ("mul", "product"):
        result = c.add(info["key"], "mul", *inputs)
        log.info(f"Add {repr(result)} using .add_product()")
    else:
        # The resulting key
        key = info["key"]
        key = key if Key.bare_name(key) else Key(key)

        # Infer the dimensions of the resulting key if ":*:" is given for the dims
        if isinstance(key, Key) and key.dims == ("*",):
            key = Key.product(key.name, *inputs, tag=key.tag)
            # log.debug(f"Inferred dimensions ({', '.join(key.dims)}) for '*'")

        # If info["comp"] is None, the task is a list that collects other keys
        _seq: Type = list
        task = []

        if info["comp"] is not None:
            _seq = tuple  # Task is a computation
            # Retrieve the function for the computation
            f = c.get_comp(info["comp"])
            if f is None:
                raise ValueError(info["comp"])
            task = [partial(f, **info.get("args", {}))]

            log.info(f"Add {repr(key)} using {f.__module__}.{f.__name__}(…)")

        task.extend(inputs)

        added = c.add(key, _seq(task), strict=True, sums=info.get("sums", False))

        if isinstance(added, tuple):
            log.info(f"    + {len(added)-1} partial sums")


@handles("report")
def report(c: Computer, info):
    """Handle one entry from the ``report:`` config section."""
    log.info(f"Add report {info['key']} with {len(info['members'])} table(s)")

    # Concatenate pyam data structures
    c.add(info["key"], tuple([c.get_comp("concat")] + info["members"]), strict=True)


@handles("units", iterate=False)
def units(c: Computer, info):
    """Handle the ``units:`` config section."""

    # Define units
    registry = pint.get_application_registry()
    try:
        defs = info["define"].strip()
        registry.define(defs)
    except KeyError:
        pass
    except (TypeError, pint.DefinitionSyntaxError, pint.RedefinitionError) as e:
        log.warning(e)
    else:
        log.info(f"Apply global unit definitions {defs}")

    # Add replacements
    for old, new in info.get("replace", {}).items():
        log.info(f"Replace unit {repr(old)} with {repr(new)}")
        REPLACE_UNITS[old] = new



"""Compatibility code for :mod:`pint`.

Notes:

- In unit expressions that contain multiple errors (e.g. undefined units *and* invalid
  syntax), the exact exception raised by pint can sometimes vary between releases.
- In pint 0.17, DefinitionSyntaxError is a subclass of SyntaxError.
  In pint 0.20, it is a subclass of ValueError.
- In pint <0.22, certain expressions raise DefinitionSyntaxError.
  In pint ≥0.22, they instead raise AssertionError.
"""
from importlib.metadata import version
from typing import Tuple, Type

import pint

try:
    PintError: Tuple[Type[Exception], ...] = (pint.PintError,)
    ApplicationRegistry: Type = pint.ApplicationRegistry
except AttributeError:  # pragma: no cover
    # Older versions of pint, e.g. 0.17
    PintError = (type("PintError", (Exception,), {}),)
    ApplicationRegistry = pint.UnitRegistry

if version("Pint") >= "0.22":
    PintError = PintError + (AssertionError,)

__all__ = [
    "ApplicationRegistry",
    "PintError",
]



import logging
import re

import pandas as pd
import pint
import pytest
from dask.core import quote

from genno import Key, Quantity
from genno.compat.pint import PintError
from genno.testing import assert_logs
from genno.util import (
    clean_units,
    collect_units,
    filter_concat_args,
    parse_units,
    partial_split,
    unquote,
)


@pytest.mark.parametrize("input, exp", (("[kg]", "kg"), ("%", "percent")))
def test_clean_units(input, exp):
    assert exp == clean_units(input)


def test_collect_units(caplog, ureg):
    q1 = Quantity(pd.Series([42, 43]), units="kg")
    # Force string units
    q1.attrs["_unit"] = "kg"

    # Units are converted to pint.Unit
    assert (ureg.kg,) == collect_units(q1)

    q2 = Quantity(pd.Series([42, 43]), name="foo")
    with caplog.at_level(logging.DEBUG):
        assert (ureg.dimensionless,) == collect_units(q2)
    assert (
        "AttrSeries 'foo' ('dim_0',) lacks units; assume dimensionless"
        == caplog.messages[-1]
    )

    assert (ureg.dimensionless,) == collect_units(1.0)

    # with pytest.raises(FileNotFoundError):
    collect_units(object())


def test_filter_concat_args(caplog):
    with assert_logs(
        caplog,
        [
            "concat() argument 'key1' missing; will be omitted",
            "concat() argument <foo:x-y-z> missing; will be omitted",
        ],
    ):
        result = list(
            filter_concat_args(
                ["key1", Quantity(pd.Series([42, 43]), units="kg"), Key("foo", "xyz")]
            )
        )

    assert len(result) == 1


msg = "unit '{}' cannot be parsed; contains invalid character(s) '{}'"


@pytest.mark.parametrize(
    "input, expected",
    (
        # Mixed units
        (["kg", "km"], (ValueError, re.escape("mixed units ['kg', 'km']"))),
        (["kg", "kg"], "kg"),
        # Units with / are defined
        (["foo/bar"], "foo/bar"),
        # Dimensionless
        ([], "dimensionless"),
        # Invalid characters, alone or with prefix
        # NB match=re.escape(msg.format("_?", "?") in pint 0.20, but varies in pint 0.17
        (["_?"], ((ValueError, pint.UndefinedUnitError), None)),
        # NB match=re.escape(msg.format("E$", "$") in pint 0.20, but varies in pint 0.17
        (["E$"], ((ValueError, pint.UndefinedUnitError), None)),
        (["kg-km"], (ValueError, re.escape(msg.format("kg-km", "-")))),
    ),
    ids=lambda argvalue: repr(argvalue),
)
def test_parse_units0(ureg, input, expected):
    if isinstance(expected, str):
        # Expected to work
        result = parse_units(input, ureg)
        assert ureg.parse_units(expected) == result
    else:
        # Expected to raise an exception
        with pytest.raises(expected[0], match=expected[1]):
            parse_units(pd.Series(input))


def test_parse_units1(ureg, caplog):
    """Multiple attempts to (re)define new units."""
    parse_units("JPY")
    parse_units("GBP/JPY")

    with pytest.raises(PintError, match="cannot be parsed; contains invalid character"):
        parse_units("GBP/JPY/$?")


def test_partial_split():
    # Function with ordinary arguments
    def func1(arg1, arg2, foo=-1, bar=-1):
        pass  # pragma: no cover

    # Function with keyword-only arguments
    def func2(arg1, arg2, *, foo, bar):
        pass  # pragma: no cover

    kwargs = dict(arg1=0, foo=1, bar=2, baz=3)

    _, extra = partial_split(func1, kwargs)
    assert {"baz"} == set(extra.keys())

    _, extra = partial_split(func2, kwargs)
    assert {"baz"} == set(extra.keys())

    with pytest.raises(TypeError):
        partial_split(1.2, kwargs)


@pytest.mark.parametrize(
    "value, exp",
    (
        # Quotable values are unwrapped
        (quote(dict(foo="bar")), dict(foo="bar")),
        (quote(["hello", "world"]), ["hello", "world"]),
        # No effect on others
        (42.0, 42.0),
    ),
)
def test_unquote(value, exp):
    assert exp == unquote(value)



"""Compatibility with :mod:`xarray`."""
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Hashable,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np
import pandas as pd
import xarray
from xarray.core.types import InterpOptions
from xarray.core.utils import is_scalar

from genno.core.types import Dims

T = TypeVar("T", covariant=True)

__all__ = [
    "DataArrayLike",
    "is_scalar",
]


class DataArrayLike(Generic[T]):
    """Class with :class:`.xarray.DataArray` -like API.

    This class is used to set signatures and types for methods and attributes on the
    generic :class:`.Quantity` class. :class:`.SparseDataArray` inherits from both this
    class and :class:`~xarray.DataArray`, and thus DataArray supplies implementations of
    these methods. In :class:`.AttrSeries`, the methods are implemented directly.
    """

    # To silence a warning in xarray
    __slots__: Tuple[str, ...] = tuple()

    # Type hints for mypy in downstream applications
    def __len__(self) -> int:
        return NotImplemented

    def __mul__(self, other):  # TODO set the return type
        ...

    def __pow__(self, other):  # TODO set the return type
        ...

    def __radd__(self, other):
        ...

    def __rmul__(self, other):
        ...

    def __rsub__(self, other):
        ...

    def __rtruediv__(self, other):
        ...

    def __truediv__(self, other):  # TODO set the return type
        ...

    def __neg__(self):
        ...

    @property
    def attrs(self) -> Dict[Any, Any]:
        return NotImplemented

    @property
    def data(self) -> Any:
        return NotImplemented

    @property
    def coords(
        self,
    ) -> xarray.core.coordinates.DataArrayCoordinates:
        return NotImplemented

    @property
    def dims(self) -> Tuple[Hashable, ...]:
        return NotImplemented

    @property
    def shape(self) -> Tuple[int, ...]:
        return NotImplemented

    @property
    def size(self) -> int:
        return NotImplemented

    def assign_coords(
        self,
        coords: Optional[Mapping[Any, Any]] = None,
        **coords_kwargs: Any,
    ):
        ...

    def astype(
        self,
        dtype,
        *,
        order=None,
        casting=None,
        subok=None,
        copy=None,
        keep_attrs=True,
    ):
        ...

    def bfill(
        self,
        dim: Hashable,
        limit: Optional[int] = None,
    ):  # TODO set the return type
        ...

    def copy(
        self,
        deep: bool = True,
        data: Any = None,
    ):  # TODO set the return type
        ...

    def cumprod(
        self,
        dim: Dims = None,
        *,
        skipna: Optional[bool] = None,
        keep_attrs: Optional[bool] = None,
        **kwargs: Any,
    ):  # TODO set the return type
        ...

    def drop_vars(
        self,
        names: Union[Hashable, Iterable[Hashable]],
        *,
        errors="raise",
    ):
        ...

    def expand_dims(
        self,
        dim=None,
        axis=None,
        **dim_kwargs: Any,
    ):  # TODO set the return type
        ...

    def ffill(
        self,
        dim: Hashable,
        limit: Optional[int] = None,
    ):  # TODO set the return type
        return NotImplemented

    def groupby(
        self,
        group,
        squeeze: bool = True,
        restore_coord_dims: bool = False,
    ):
        ...

    def interp(
        self,
        coords: Optional[Mapping[Any, Any]] = None,
        method: InterpOptions = "linear",
        assume_sorted: bool = False,
        kwargs: Optional[Mapping[str, Any]] = None,
        **coords_kwargs: Any,
    ):
        ...

    def item(self, *args):
        ...

    def pipe(
        self,
        func: Union[Callable[..., T], Tuple[Callable[..., T], str]],
        *args: Any,
        **kwargs: Any,
    ) -> T:
        return NotImplemented

    def rename(
        self,
        new_name_or_name_dict: Union[Hashable, Mapping[Any, Hashable]] = None,
        **names: Hashable,
    ):
        ...

    def round(self, *args, **kwargs):
        ...

    def sel(
        self,
        indexers: Optional[Mapping[Any, Any]] = None,
        method: Optional[str] = None,
        tolerance=None,
        drop: bool = False,
        **indexers_kwargs: Any,
    ) -> T:
        return NotImplemented

    def shift(
        self,
        shifts: Optional[Mapping[Any, int]] = None,
        fill_value: Any = None,
        **shifts_kwargs: int,
    ):
        ...

    def sum(
        self,
        dim: Dims = None,
        # Signature from xarray.DataArray
        *,
        skipna: Optional[bool] = None,
        min_count: Optional[int] = None,
        keep_attrs: Optional[bool] = None,
        **kwargs: Any,
    ):
        ...

    def to_dataframe(
        self,
        name: Optional[Hashable] = None,
        dim_order: Optional[Sequence[Hashable]] = None,
    ) -> pd.DataFrame:
        ...

    def to_numpy(self) -> np.ndarray:
        return NotImplemented

    def to_series(self) -> pd.Series:
        """Like :meth:`xarray.DataArray.to_series`."""
        # Provided only for type-checking in other packages. AttrSeries implements;
        # SparseDataArray uses the xr.DataArray method.
        return NotImplemented



import pytest

from genno import Computer, Operator


def null():
    pass  # pragma: no cover


class TestOperator:
    def test_add_tasks(self):
        op = Operator.define(null)

        c = Computer()
        with pytest.raises(NotImplementedError):
            op.add_tasks(c)



import re
from os import PathLike
from typing import Mapping, MutableMapping, Optional, Set, Union

import graphviz
from dask.core import get_dependencies, ishashable, istask
from dask.dot import graphviz_to_file, name

from genno.core.describe import is_list_of_keys, label


def key_label(key):
    return unwrap(str(key))


_UNWRAP_EXPR = re.compile("^<(.*)>$")


def unwrap(label: str) -> str:
    """Unwrap any number of paired '<' and '>' at the start/end of `label`.

    These characters cause errors in graphviz/dot.
    """
    while True:
        result = _UNWRAP_EXPR.sub(r"\1", label)
        if result == label:
            return result
        else:
            label = result


def visualize(
    dsk: Mapping,
    filename: Optional[Union[str, PathLike]] = None,
    format: Optional[str] = None,
    data_attributes: Optional[Mapping] = None,
    function_attributes: Optional[Mapping] = None,
    graph_attr: Optional[Mapping] = None,
    node_attr: Optional[Mapping] = None,
    edge_attr: Optional[Mapping] = None,
    collapse_outputs=False,
    **kwargs,
):
    """Generate a Graphviz visualization of `dsk`.

    This is merged and extended version of :func:`dask.base.visualize`,
    :func:`dask.dot.dot_graph`, and :func:`dask.dot.to_graphviz` that produces output
    that is informative for genno graphs.

    Parameters
    ----------
    dsk :
        The graph to display.
    filename : Path or str, optional
        The name of the file to write to disk. If the file name does not have a suffix,
        ".png" is used by default. If `filename` is :data:`None`, no file is written,
        and dask communicates with :program:`dot` using only pipes.
    format : {'png', 'pdf', 'dot', 'svg', 'jpeg', 'jpg'}, optional
        Format in which to write output file, if not given by the suffix of `filename`.
        Default "png".
    data_attributes :
        Graphviz attributes to apply to single nodes representing keys, in addition to
        `node_attr`.
    function_attributes :
        Graphviz attributes to apply to single nodes representing operations or
        functions, in addition to `node_attr`.
    graph_attr :
        Mapping of (attribute, value) pairs for the graph. Passed directly to
        :class:`.graphviz.Digraph`.
    node_attr :
        Mapping of (attribute, value) pairs set for all nodes. Passed directly to
        :class:`.graphviz.Digraph`.
    edge_attr :
        Mapping of (attribute, value) pairs set for all edges. Passed directly to
        :class:`.graphviz.Digraph`.
    collapse_outputs : bool, optional
        Omit nodes for keys that are the output of intermediate calculations.
    kwargs :
        All other keyword arguments are added to `graph_attr`.

    Examples
    --------

    .. _visualize-example:

    Prepare a computer:

    >>> from genno import Computer
    >>> from genno.testing import add_test_data
    >>> c = Computer()
    >>> add_test_data(c)
    >>> c.add_product("z", "x:t", "x:y")
    >>> c.add("y::0", itemgetter(0), "y")
    >>> c.add("y0", "y::0")
    >>> c.add("index_to", "z::indexed", "z:y", "y::0")
    >>> c.add_single("all", ["z::indexed", "t", "config", "x:t"])

    Visualize its contents:

    >>> c.visualize("example.svg")

    This produces the output:

    .. image:: _static/visualize.svg
       :alt: Example output from graphviz.visualize.

    See also
    --------
    .describe.label
    """
    # Handle arguments
    item_attr = {
        "data": data_attributes or {},
        "func": function_attributes or {},
    }
    _graph_attr: MutableMapping = dict(graph_attr) if graph_attr else {}
    _node_attr: MutableMapping = dict(node_attr) if node_attr else {}
    edge_attr = edge_attr or {}

    # Default attributes
    _graph_attr.setdefault("rankdir", "BT")
    _node_attr.setdefault("fontname", "helvetica")

    # Assume unused kwargs are for graph_attr
    _graph_attr.update(kwargs)

    # Use a directional shape like [> in LR mode; otherwise a box
    key_shape = "cds" if _graph_attr["rankdir"] == "LR" else "box"

    g = graphviz.Digraph(
        graph_attr=_graph_attr, node_attr=_node_attr, edge_attr=edge_attr
    )

    seen = set()  # Nodes or edges already seen
    connected: Set[str] = set()  # Nodes already connected to the graph

    # Shorthand
    def _attrs(kind, key, **defaults):
        """Prepare a copy from `item_attr` for `kind` with `defaults`."""
        result = item_attr[kind].get(key, {}).copy()
        for k, v in defaults.items():
            result.setdefault(k, v)
        return result

    def _edge(a, b):
        """Add an edge to `g` and update `connected`."""
        g.edge(a, b)
        connected.update(a, b)

    # Iterate over keys, tasks in the graph
    for k, v in dsk.items():
        # A unique "name" for the node within `g`; similar to hash(k).
        k_name = name(k)

        if istask(v):
            # A task

            # Node name for the operation
            func_name = name((k, "function")) if not collapse_outputs else k_name

            # Add a node for the operation
            if collapse_outputs or func_name not in seen:
                seen.add(func_name)
                attrs = _attrs(
                    "func", k, label=unwrap(label(v[0], max_length=50)), shape=key_shape
                )
                g.node(func_name, **attrs)

            # Add an edge between the operation-node and the key-node of its output
            if not collapse_outputs:
                _edge(func_name, k_name)

            # Add edges between the operation-node and the key-nodes for each of its
            # inputs
            for dep in get_dependencies(dsk, k):
                dep_name = name(dep)
                if dep_name not in seen:
                    seen.add(dep_name)
                    attrs = _attrs("data", dep, label=key_label(dep), shape="ellipse")
                    g.node(dep_name, **attrs)
                _edge(dep_name, func_name)

        elif ishashable(v) and v in dsk:
            # Simple alias of k → v
            _edge(name(v), k_name)

        elif is_list_of_keys(v, dsk):
            # k is a list of multiple keys (genno extension)
            for _v in v:
                _edge(name(_v), k_name)

        if (not collapse_outputs or k_name in connected) and k_name not in seen:
            # Something else that hasn't been seen: add a node that may never be
            # connected
            seen.add(k_name)
            attrs = _attrs("data", k, label=key_label(k), shape="ellipse")
            g.node(k_name, **attrs)

    return graphviz_to_file(g, None if filename is None else str(filename), format)



from typing import TYPE_CHECKING, Iterable, List, Mapping, Optional, Union

if TYPE_CHECKING:
    from sdmx.model.common import Code, Codelist


def codelist_to_groups(
    codes: Union["Codelist", Iterable["Code"]], dim: Optional[str] = None
) -> Mapping[str, Mapping[str, List[str]]]:
    """Convert `codes` into a mapping from parent items to their children.

    The returned value is suitable for use with :func:`~.computations.aggregate`.

    Parameters
    ----------
    codes
        Either a :class:`sdmx.Codelist <sdmx.model.common.Codelist>` object or any
        iterable of :class:`sdmx.Code <sdmx.model.common.Code>`.
    dim : str, optional
        Dimension to aggregate. If `codes` is a code list and `dim` is not given, the
        ID of the code list is used; otherwise `dim` must be supplied.
    """
    from sdmx.model.common import Codelist

    if isinstance(codes, Codelist):
        items: Iterable["Code"] = codes.items.values()
        dim = dim or codes.id
    else:
        items = codes

    if dim is None:
        raise ValueError("Must provide a dimension ID for aggregation")

    groups = dict()
    for code in filter(lambda c: len(c.child), items):
        groups[code.id] = list(map(str, code.child))

    return {dim: groups}



from typing import Any, Dict, Hashable, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd

try:
    import sparse

    HAS_SPARSE = True
except ImportError:
    HAS_SPARSE = False
import xarray as xr
from xarray.core import dtypes
from xarray.core.utils import either_dict_or_kwargs

from genno.core.quantity import Quantity


@xr.register_dataarray_accessor("_sda")
class SparseAccessor:
    """:mod:`xarray` accessor to help :class:`SparseDataArray`.

    See the xarray accessor documentation, e.g.
    :func:`~xarray.register_dataarray_accessor`.
    """

    def __init__(self, obj):
        self.da = obj

    def convert(self):
        """Return a :class:`SparseDataArray` instance."""
        if not self.da._sda.COO_data:
            # Dense (numpy.ndarray) data; convert to sparse
            data = sparse.COO.from_numpy(self.da.data, fill_value=np.nan)
        elif not np.isnan(self.da.data.fill_value):
            # sparse.COO with non-NaN fill value; copy and change
            data = self.da.data.copy(deep=False)
            data.fill_value = data.dtype.type(np.nan)
        else:
            # No change
            data = self.da.data

        if isinstance(self.da, SparseDataArray):
            # Replace the variable, returning a copy
            variable = self.da.variable._replace(data=data)
            return self.da._replace(variable=variable)
        else:
            # Construct
            return SparseDataArray(
                data=data,
                coords=self.da.coords,
                dims=self.da.dims,
                name=self.da.name,
                attrs=self.da.attrs,
            )

    @property
    def COO_data(self):
        """:obj:`True` if the DataArray has :class:`sparse.COO` data."""
        return isinstance(self.da.data, sparse.COO)

    @property
    def dense(self):
        """Return a copy with dense (:class:`.ndarray`) data."""
        # Use existing method xr.Variable._to_dense()
        return self.da._replace(variable=self.da.variable._to_dense())

    @property
    def dense_super(self):
        """Return a proxy to a :class:`.ndarray`-backed :class:`.DataArray`."""
        return super(SparseDataArray, self.dense)


class OverrideItem:
    """Override :meth:`xarray.DataArray.item`.

    The :meth:`item` method is set dynamically by
    :class:`xarray.ops.IncludeNumpySameMethods`, a parent of
    :class:`xarray.arithmetic.DataArrayArithmetic` and thus of DataArray.
    That has the effect of overriding an ordinary :meth:`item` method defined on
    :class:`SparseDataArray`.

    This class, placed higher in the MRO for SparseDataArray, cancels out that effect.
    """

    __slots__ = ()

    def __init_subclass__(cls, **kwargs):
        setattr(cls, "item", cls._item)


class SparseDataArray(OverrideItem, xr.DataArray, Quantity):
    """:class:`~xarray.DataArray` with sparse data.

    SparseDataArray uses :class:`sparse.COO` for storage with :data:`numpy.nan`
    as its :attr:`sparse.COO.fill_value`. Some methods of :class:`~xarray.DataArray`
    are overridden to ensure data is in sparse, or dense, format as necessary, to
    provide expected functionality not currently supported by :mod:`sparse`, and to
    avoid exhausting memory for some operations that require dense data.
    """

    __slots__: Tuple[str, ...] = tuple()

    def __init__(
        self,
        data: Any = dtypes.NA,
        coords: Union[Sequence[Tuple], Mapping[Hashable, Any], None] = None,
        dims: Union[Hashable, Sequence[Hashable], None] = None,
        name: Hashable = None,
        attrs: Optional[Mapping] = None,
        # internal parameters
        indexes: Optional[Dict[Hashable, pd.Index]] = None,
        fastpath: bool = False,
        **kwargs,
    ):
        if fastpath:
            return xr.DataArray.__init__(
                self, data, coords, dims, name, attrs, indexes, fastpath
            )

        attrs = Quantity._collect_attrs(data, attrs, kwargs)

        assert 0 == len(
            kwargs
        ), f"Unrecognized kwargs {kwargs.keys()} to SparseDataArray()"

        if isinstance(data, int):
            data = float(data)

        data, name = Quantity._single_column_df(data, name)

        if isinstance(data, pd.Series):
            # Possibly converted from pd.DataFrame, above
            if data.dtype == int:
                # Ensure float data
                data = data.astype(float)
            data = xr.DataArray.from_series(data, sparse=True)

        if isinstance(data, xr.DataArray):
            # Possibly converted from pd.Series, above
            coords = data._coords
            data = data.variable

        # Invoke the xr.DataArray constructor
        xr.DataArray.__init__(self, data, coords, dims, name, attrs)

        if not isinstance(self.variable.data, sparse.COO):
            # Dense (numpy.ndarray) data; convert to sparse
            data = sparse.COO.from_numpy(self.variable.data, fill_value=np.nan)
        elif not np.isnan(self.variable.data.fill_value):
            # sparse.COO with non-NaN fill value; copy and change
            data = self.variable.data.copy(deep=False)
            data.fill_value = data.dtype.type(np.nan)
        else:
            # No change
            return

        # Replace the variable
        self._variable = self._variable._replace(data=data)

    @classmethod
    def from_series(cls, obj, sparse=True):
        """Convert a pandas.Series into a SparseDataArray."""
        # Call the parent method always with sparse=True, then re-wrap
        return xr.DataArray.from_series(obj, sparse=True)._sda.convert()

    def ffill(self, dim: Hashable, limit: Optional[int] = None):
        """Override :meth:`~xarray.DataArray.ffill` to auto-densify."""
        return self._sda.dense_super.ffill(dim, limit)._sda.convert()

    def _item(self, *args):
        """Like :meth:`~xarray.DataArray.item`."""
        # See OverrideItem
        if len(args):  # pragma: no cover
            super().item(*args)
        elif len(self.data.shape) == 0:
            return (
                self.data.data[0]
                if isinstance(self.data, sparse.COO)
                else self.data.item()
            )
        else:
            raise ValueError("can only convert an array of size 1 to a Python scalar")

    def sel(
        self,
        indexers: Optional[Mapping[Any, Any]] = None,
        method: Optional[str] = None,
        tolerance=None,
        drop: bool = False,
        **indexers_kwargs: Any,
    ) -> "SparseDataArray":
        """Return a new array by selecting labels along the specified dim(s).

        Overrides :meth:`~xarray.DataArray.sel` to handle >1-D indexers with sparse
        data.
        """
        indexers = either_dict_or_kwargs(indexers, indexers_kwargs, "sel")
        if isinstance(indexers, dict) and len(indexers) > 1:
            result = self
            for k, v in indexers.items():
                result = result.sel(
                    {k: v}, method=method, tolerance=tolerance, drop=drop
                )
            return result
        else:
            return (
                super()
                .sel(indexers=indexers, method=method, tolerance=tolerance, drop=drop)
                ._sda.convert()
            )

    def to_dataframe(
        self,
        name: Optional[Hashable] = None,
        dim_order: Optional[Sequence[Hashable]] = None,
    ) -> pd.DataFrame:
        """Convert this array and its coords into a :class:`~xarray.DataFrame`.

        Overrides :meth:`~xarray.DataArray.to_dataframe`.
        """
        if dim_order is not None:
            raise NotImplementedError("dim_order arg to to_dataframe()")
        return self.to_series().to_frame(name or self.name or "value")

    def to_series(self) -> pd.Series:
        """Convert this array into a :class:`~pandas.Series`.

        Overrides :meth:`~xarray.DataArray.to_series` to create the series without
        first converting to a potentially very large :class:`numpy.ndarray`.
        """
        # Use SparseArray.coords and .data (each already 1-D) to construct the pd.Series

        # Construct a pd.MultiIndex without using .from_product
        index = pd.MultiIndex.from_arrays(self.data.coords, names=self.dims).set_levels(
            [self.coords[d].values for d in self.dims]
        )
        return pd.Series(self.data.data, index=index, name=self.name)



import logging
import warnings
from functools import partial
from itertools import tee
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Hashable,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

if TYPE_CHECKING:  # pragma: no cover
    from _typeshed import SupportsRichComparisonT

import pandas as pd
import pandas.core.indexes.base as ibase
import xarray as xr
from xarray.core.coordinates import Coordinates
from xarray.core.indexes import Indexes
from xarray.core.utils import either_dict_or_kwargs

from genno.compat.xarray import is_scalar

from .quantity import Quantity, possible_scalar
from .types import Dims

log = logging.getLogger(__name__)


def _binop(name: str, swap: bool = False):
    def method(self, other):
        other = possible_scalar(other)

        # For __r*__ methods
        a, b = (other, self) if swap else (self, other)

        # Ensure both operands are multi-indexed, and have at least 1 common dim
        if a.dims:
            left = a
            order, right = b.align_levels(left)
        else:
            right = b
            order, left = a.align_levels(right)

        return getattr(left, name)(right).dropna().reorder_levels(order)

    return method


class AttrSeriesCoordinates(Coordinates):
    def __init__(self, obj):
        self._data = obj
        self._idx = obj.index.remove_unused_levels()

    @property
    def _names(self):
        return tuple(filter(None, self._idx.names))

    @property
    def variables(self):
        result = {}
        for name, levels in zip(self._idx.names, self._idx.levels):
            if name is None:
                continue
            result[name] = levels.unique()
        return result

    def __contains__(self, key: Hashable) -> bool:
        return key in self._names

    def __getitem__(self, key):
        levels = self._idx.levels[self._idx.names.index(key)].to_list()
        return xr.DataArray(levels, coords={key: levels})


class AttrSeries(pd.Series, Quantity):
    """:class:`pandas.Series` subclass imitating :class:`xarray.DataArray`.

    The AttrSeries class provides similar methods and behaviour to
    :class:`xarray.DataArray`, so that :mod:`genno.computations` functions and user
    code can use xarray-like syntax. In particular, this allows such code to be agnostic
    about the order of dimensions.

    Parameters
    ----------
    units : str or pint.Unit, optional
        Set the units attribute. The value is converted to :class:`pint.Unit` and added
        to `attrs`.
    attrs : :class:`~collections.abc.Mapping`, optional
        Set the :attr:`~pandas.Series.attrs` of the AttrSeries. This attribute was added
        in `pandas 1.0 <https://pandas.pydata.org/docs/whatsnew/v1.0.0.html>`_, but is
        not currently supported by the Series constructor.
    """

    # See https://pandas.pydata.org/docs/development/extending.html
    @property
    def _constructor(self):
        return AttrSeries

    def __init__(self, data=None, *args, name=None, attrs=None, **kwargs):
        attrs = Quantity._collect_attrs(data, attrs, kwargs)

        if isinstance(data, (pd.Series, xr.DataArray)):
            # Extract name from existing object or use the argument
            name = ibase.maybe_extract_name(name, data, type(self))

            try:
                # Pre-convert to pd.Series from xr.DataArray to preserve names and
                # labels. For AttrSeries, this is a no-op (see below).
                data = data.to_series()
            except AttributeError:
                # pd.Series
                pass
            except ValueError:
                # xr.DataArray
                if data.shape == tuple():
                    # data is a scalar/0-dimensional xr.DataArray. Pass the 1 value
                    data = data.data
                else:  # pragma: no cover
                    raise
            else:
                attrs.update()

        data, name = Quantity._single_column_df(data, name)

        if data is None:
            kwargs["dtype"] = float

        # Don't pass attrs to pd.Series constructor; it currently does not accept them
        pd.Series.__init__(self, data, *args, name=name, **kwargs)

        # Ensure a MultiIndex
        try:
            self.index.levels
        except AttributeError:
            # Assign the dimension name "dim_0" if 1-D with no names
            kw = {}
            if len(self.index) > 1 and self.index.name is None:
                kw["names"] = ["dim_0"]
            self.index = pd.MultiIndex.from_product([self.index], **kw)

        # Update the attrs after initialization
        self.attrs.update(attrs)

    # Binary operations
    __mul__ = _binop("mul")
    __pow__ = _binop("pow")
    __rtruediv__ = _binop("div", swap=True)
    __truediv__ = _binop("div")

    def __repr__(self):
        return (
            super().__repr__() + f", units: {self.attrs.get('_unit', 'dimensionless')}"
        )

    @classmethod
    def from_series(cls, series, sparse=None):
        """Like :meth:`xarray.DataArray.from_series`."""
        return AttrSeries(series)

    def assign_coords(self, coords=None, **coord_kwargs):
        """Like :meth:`xarray.DataArray.assign_coords`."""
        coords = either_dict_or_kwargs(coords, coord_kwargs, "assign_coords")

        # Construct a new index
        new_idx = self.index.copy()
        for dim, values in coords.items():
            expected_len = len(self.index.levels[self.index.names.index(dim)])
            if expected_len != len(values):
                raise ValueError(
                    f"conflicting sizes for dimension {repr(dim)}: length "
                    f"{expected_len} on <this-array> and length {len(values)} on "
                    f"{repr(dim)}"
                )

            new_idx = new_idx.set_levels(values, level=dim)

        # Return a new object with the new index
        return self.set_axis(new_idx)

    def bfill(self, dim: Hashable, limit: Optional[int] = None):
        """Like :meth:`xarray.DataArray.bfill`."""
        # TODO this likely does not work for 1-D quantities due to unstack(); test and
        #      if needed use _maybe_groupby()
        return self._replace(
            self.unstack(dim)
            .bfill(axis=1, limit=limit)
            .stack()
            .reorder_levels(self.dims),
        )

    @property
    def coords(self):
        """Like :attr:`xarray.DataArray.coords`. Read-only."""
        return AttrSeriesCoordinates(self)

    def cumprod(self, dim=None, axis=None, skipna=None, **kwargs):
        """Like :meth:`xarray.DataArray.cumprod`."""
        if axis:
            log.info(f"{self.__class__.__name__}.cumprod(…, axis=…) is ignored")
        if skipna is None:
            skipna = self.dtype == float
        if dim in (None, "..."):
            if len(self.dims) > 1:
                raise NotImplementedError("cumprod() over multiple dimensions")
            dim = self.dims[0]

        def _(s):
            # Invoke cumprod from the parent class pd.Series
            return super(pd.Series, s).cumprod(skipna=skipna, **kwargs)

        return self._replace(self._groupby_apply(dim, sorted(self.coords[dim].data), _))

    @property
    def data(self):
        return self.values

    @property
    def dims(self) -> Tuple[Hashable, ...]:
        """Like :attr:`xarray.DataArray.dims`."""
        # If 0-D, the single dimension has name `None` → discard
        return tuple(filter(None, self.index.names))

    @property
    def shape(self) -> Tuple[int, ...]:
        """Like :attr:`xarray.DataArray.shape`."""
        idx = self.index.remove_unused_levels()
        return tuple(len(idx.levels[i]) for i in map(idx.names.index, self.dims))

    def drop(self, label):
        """Like :meth:`xarray.DataArray.drop`."""
        return self.droplevel(label)

    def drop_vars(
        self, names: Union[Hashable, Iterable[Hashable]], *, errors: str = "raise"
    ):
        """Like :meth:`xarray.DataArray.drop_vars`."""

        return self.droplevel(names)

    def expand_dims(self, dim=None, axis=None, **dim_kwargs: Any) -> "AttrSeries":
        """Like :meth:`xarray.DataArray.expand_dims`."""
        if isinstance(dim, list):
            dim = dict.fromkeys(dim, [])
        dim = either_dict_or_kwargs(dim, dim_kwargs, "expand_dims")
        if axis is not None:
            raise NotImplementedError  # pragma: no cover

        result = self
        for name, values in reversed(list(dim.items())):
            N = len(values)
            if N == 0:  # Dimension without labels
                N, values = 1, [None]
            result = pd.concat([result] * N, keys=values, names=[name])

        # Ensure `result` is multiindexed
        try:
            i = result.index.names.index(None)
        except ValueError:
            pass
        else:
            assert 2 == len(result.index.names)
            result.index = pd.MultiIndex.from_product([result.index.droplevel(i)])

        return result

    def ffill(self, dim: Hashable, limit: Optional[int] = None):
        """Like :meth:`xarray.DataArray.ffill`."""
        # TODO this likely does not work for 1-D quantities due to unstack(); test and
        #      if needed use _maybe_groupby()
        return self._replace(
            self.unstack(dim)
            .ffill(axis=1, limit=limit)
            .stack()
            .reorder_levels(self.dims),
        )

    def item(self, *args):
        """Like :meth:`xarray.DataArray.item`."""
        if len(args) and args != (None,):
            raise NotImplementedError
        elif self.size != 1:
            raise ValueError
        return self.iloc[0]

    def interp(
        self,
        coords: Optional[Mapping[Hashable, Any]] = None,
        method: str = "linear",
        assume_sorted: bool = True,
        kwargs: Optional[Mapping[str, Any]] = None,
        **coords_kwargs: Any,
    ):
        """Like :meth:`xarray.DataArray.interp`.

        This method works around two long-standing bugs in :mod:`pandas`:

        - `pandas-dev/pandas#25460 <https://github.com/pandas-dev/pandas/issues/25460>`_
        - `pandas-dev/pandas#31949 <https://github.com/pandas-dev/pandas/issues/31949>`_
        """
        from scipy.interpolate import interp1d

        if kwargs is None:
            kwargs = {}

        coords = either_dict_or_kwargs(coords, coords_kwargs, "interp")
        if len(coords) > 1:
            raise NotImplementedError("interp() on more than 1 dimension")

        # Unpack the dimension and levels (possibly overlapping with existing)
        dim = list(coords.keys())[0]
        levels = coords[dim]
        # Ensure a list
        if isinstance(levels, (int, float)):
            levels = [levels]

        def _flat_index(obj: AttrSeries):
            """Unpack a 1-D MultiIndex from an AttrSeries."""
            return [v[0] for v in obj.index]

        # Group by `dim` so that each level appears ≤ 1 time in `group_series`

        def _(s):
            # Work around https://github.com/pandas-dev/pandas/issues/31949
            # Location of existing values
            x = s.notna()

            # Create an interpolator from the existing values
            i = interp1d(_flat_index(s[x]), s[x], kind=method, **kwargs)

            return s.fillna(pd.Series(i(_flat_index(s[~x])), index=s[~x].index))

        result = self._groupby_apply(dim, levels, _)

        # - Restore dimension order and attributes.
        # - Select only the desired `coords`.
        return self._replace(result.reorder_levels(self.dims)).sel(coords)

    def rename(
        self,
        new_name_or_name_dict: Union[Hashable, Mapping[Hashable, Hashable]] = None,
        **names: Hashable,
    ):
        """Like :meth:`xarray.DataArray.rename`."""
        if new_name_or_name_dict is None or isinstance(new_name_or_name_dict, Mapping):
            index = either_dict_or_kwargs(new_name_or_name_dict, names, "rename")
            return self.rename_axis(index=index)
        else:
            assert 0 == len(names)
            return super().rename(new_name_or_name_dict)

    def sel(
        self,
        indexers: Optional[Mapping[Any, Any]] = None,
        method: Optional[str] = None,
        tolerance=None,
        drop: bool = False,
        **indexers_kwargs: Any,
    ):
        """Like :meth:`xarray.DataArray.sel`."""
        if method is not None:
            raise NotImplementedError(f"AttrSeries.sel(…, method={method!r})")
        if tolerance is not None:
            raise NotImplementedError(f"AttrSeries.sel(…, tolerance={tolerance!r})")

        indexers = either_dict_or_kwargs(indexers, indexers_kwargs, "sel")

        if len(indexers) == 1:
            level, key = list(indexers.items())[0]
            if isinstance(key, str) and not drop:
                # When using .loc[] to select 1 label on 1 level, pandas drops the
                # level. Use .xs() to avoid this behaviour unless drop=True
                return AttrSeries(self.xs(key, level=level, drop_level=False))

        if len(indexers) and all(
            isinstance(i, xr.DataArray) for i in indexers.values()
        ):
            # DataArray indexers

            # Combine indexers in a data set; dimensions are aligned
            ds = xr.Dataset(indexers)

            # All dimensions indexed
            dims_indexed = set(indexers.keys())
            # Dimensions to discard
            dims_drop = set(ds.data_vars.keys())

            # Check contents of indexers
            if any(ds.isnull().any().values()):
                raise IndexError(
                    f"Dimensions of indexers mismatch: {ds.notnull().sum()}"
                )
            elif len(ds.dims) > 1:
                raise NotImplementedError(  # pragma: no cover
                    f"map to > 1 dimensions {repr(ds.dims)} with AttrSeries.sel()"
                )

            # pd.Index object with names and levels of the new dimension to be created
            idx = ds.coords.to_index()

            # Dimensions to drop on sliced data to avoid duplicated dimensions
            drop_slice = list(dims_indexed - dims_drop)

            # Dictionary of Series to concatenate
            series = {}

            # Iterate over labels in the new dimension
            for label in idx:
                # Get a slice from the indexers corresponding to this label
                loc_ds = ds.sel({idx.name: label})

                # Assemble a key with one element for each dimension
                seq0 = [loc_ds.get(d) for d in self.dims]
                # Replace None from .get() with slice(None) or unpack a single value
                seq1 = [slice(None) if item is None else item.item() for item in seq0]

                # Use the key to retrieve 1+ integer locations; slice; store
                series[label] = self.iloc[self.index.get_locs(seq1)].droplevel(
                    drop_slice
                )

            # Rejoin to a single data frame; drop the source levels
            data = pd.concat(series, names=[idx.name]).droplevel(list(dims_drop))
        else:
            # Other indexers

            # Iterate over dimensions
            idx = []
            to_drop = set()
            for dim in self.dims:
                # Get an indexer for this dimension
                i = indexers.get(dim, slice(None))

                if is_scalar(i) and (i != slice(None)) and drop:
                    to_drop.add(dim)

                # Maybe unpack an xarray DataArray indexers, for pandas
                idx.append(i.data if isinstance(i, xr.DataArray) else i)

            # Silence a warning from pandas ≥1.4 that may be spurious
            # FIXME investigate, adjust the code, remove the filter
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    ".*indexing on a MultiIndex with a nested sequence.*",
                    FutureWarning,
                )
                # Select
                data = self.loc[tuple(idx)]

            # Only drop if not returning a scalar value
            if isinstance(data, pd.Series):
                # Drop levels where a single value was selected
                data = data.droplevel(list(to_drop & set(data.index.names)))

        # Return
        return self._replace(data)

    def shift(
        self,
        shifts: Optional[Mapping[Hashable, int]] = None,
        fill_value: Any = None,
        **shifts_kwargs: int,
    ):
        """Like :meth:`xarray.DataArray.shift`."""
        shifts = either_dict_or_kwargs(shifts, shifts_kwargs, "shift")

        # Apply shifts one-by-one
        result = self
        for dim, periods in shifts.items():
            levels = sorted(self.coords[dim].data)

            def _(s):
                # Invoke shift from the parent class pd.Series
                return super(AttrSeries, s).shift(
                    periods=periods, fill_value=fill_value
                )

            result = result._groupby_apply(dim, levels, _)

        return self._replace(result)

    def sum(
        self,
        dim: Dims = None,
        # Signature from xarray.DataArray
        # *,
        skipna: Optional[bool] = None,
        min_count: Optional[int] = None,
        keep_attrs: Optional[bool] = None,
        **kwargs: Any,
    ) -> "AttrSeries":
        """Like :meth:`xarray.DataArray.sum`."""
        if skipna is not None or min_count is not None:
            raise NotImplementedError

        if dim is None or isinstance(dim, Hashable):
            dim = tuple(filter(None, (dim,)))

        # Check dimensions
        bad_dims = set(dim) - set(self.index.names)
        if bad_dims:
            raise ValueError(
                f"{bad_dims} not found in array dimensions {self.index.names}"
            )

        # Create the object on which to .sum()
        return self._replace(self._maybe_groupby(dim).sum(**kwargs))

    def squeeze(self, dim=None, *args, **kwargs):
        """Like :meth:`xarray.DataArray.squeeze`."""
        assert kwargs.pop("drop", True)

        idx = self.index.remove_unused_levels()

        to_drop = []
        for i, name in enumerate(filter(None, idx.names)):
            if dim and name != dim:
                continue
            elif len(idx.levels[i]) > 1:
                if dim is None:
                    continue
                else:
                    raise ValueError(
                        "cannot select a dimension to squeeze out which has length "
                        "greater than one"
                    )

            to_drop.append(name)

        if dim and not to_drop:
            # Specified dimension does not exist
            raise KeyError(dim)

        return self.droplevel(to_drop)

    def transpose(self, *dims):
        """Like :meth:`xarray.DataArray.transpose`."""
        return self.reorder_levels(dims)

    def to_dataframe(
        self,
        name: Optional[Hashable] = None,
        dim_order: Optional[Sequence[Hashable]] = None,
    ) -> pd.DataFrame:
        """Like :meth:`xarray.DataArray.to_dataframe`."""
        if dim_order is not None:
            raise NotImplementedError("dim_order arg to to_dataframe()")

        self.name = name or self.name or "value"  # type: ignore
        return self.to_frame()

    def to_series(self):
        """Like :meth:`xarray.DataArray.to_series`."""
        return self

    @property
    def xindexes(self):  # pragma: no cover
        # NB incomplete implementation; currently sufficient that this property exists
        return Indexes(dict(), None)

    # Internal methods
    def align_levels(
        self, other: "AttrSeries"
    ) -> Tuple[Sequence[Hashable], "AttrSeries"]:
        """Return a copy of `self` with ≥1 dimension(s) in the same order as `other`.

        Work-around for https://github.com/pandas-dev/pandas/issues/25760 and other
        limitations of :class:`pandas.Series`.
        """
        # Union of dimensions of `self` and `other`; initially just other
        d_union = list(other.dims)

        # Lists of common dimensions, and dimensions on `other` missing from `self`.
        d_common = []  # Common dimensions of `self` and `other`
        d_other_only = []  # (dimension, index) of `other` missing from `self`
        for i, d in enumerate(d_union):
            if d in self.index.names:
                d_common.append(d)
            else:
                d_other_only.append((d, i))

        result = self
        d_result = []  # Order of dimensions on the result

        if len(d_common) == 0:
            # No common dimensions between `other` and `self`
            if len(d_other_only):
                # …but `other` is ≥1D
                # Broadcast the result over the final missing dimension of `other`
                d, i = d_other_only[-1]
                result = result.expand_dims({d: other.index.levels[i]})
                # Reordering starts with this dimension
                d_result.append(d)
            elif not result.dims:
                # Both `self` and `other` are scalar
                d_result.append(None)
        else:
            # Some common dimensions exist; no need to broadcast, only reorder
            d_result.extend(d_common)

        # Append the dimensions of `self`
        i1, i2 = tee(filter(lambda n: n not in other.dims, self.dims), 2)
        d_union.extend(i1)
        d_result.extend(i2)

        return d_union or [None], result.reorder_levels(d_result or [None])

    def _groupby_apply(
        self,
        dim: Hashable,
        levels: Iterable["SupportsRichComparisonT"],
        func: Callable[["AttrSeries"], "AttrSeries"],
    ) -> "AttrSeries":
        """Group along `dim`, ensure levels `levels`, and apply `func`.

        `func` should accept and return AttrSeries. The resulting AttrSeries are
        concatenated again along `dim`.
        """
        # Preserve order of dimensions
        dims = self.dims

        # Dimension other than `dim`
        d_other = list(filter(lambda d: d != dim, dims))

        def _join(base, item):
            """Rejoin a full key for the MultiIndex in the correct order."""
            # Wrap a scalar `base` (only occurs with len(other_dims) == 1; pandas < 2.0)
            base = list(base) if isinstance(base, tuple) else [base]
            return [(base[d_other.index(d)] if d in d_other else item[0]) for d in dims]

        # Grouper or iterable of (key, pd.Series)
        groups = self.groupby(d_other) if len(d_other) else [(None, self)]

        # Iterate over groups, accumulating modified series
        result = []
        for group_key, group_series in groups:
            # Work around https://github.com/pandas-dev/pandas/issues/25460; can't do:
            # group_series.reindex(…, level=dim)

            # Create 1-D MultiIndex for `dim` with the union of existing coords and
            # `levels`
            _levels = set(levels)
            _levels.update(group_series.index.get_level_values(dim))
            idx = pd.MultiIndex.from_product([sorted(_levels)], names=[dim])
            # Reassemble full MultiIndex with the new coords added along `dim`
            full_idx = pd.MultiIndex.from_tuples(
                map(partial(_join, group_key), idx), names=dims
            )

            # - Reindex with `full_idx` to insert NaNs for new `levels`.
            # - Replace the with the 1D index for `dim` only.
            # - Apply `func`.
            # - Restore the full index.
            result.append(
                func(group_series.reindex(full_idx).set_axis(idx)).set_axis(full_idx)
            )

        return pd.concat(result)

    def _maybe_groupby(self, dim):
        """Return an object for operations along dimension(s) `dim`.

        If `dim` is a subset of :attr:`dims`, returns a SeriesGroupBy object along the
        other dimensions.
        """
        if len(set(dim)) in (0, len(self.index.names)):
            return cast(pd.Series, super())
        else:
            # Group on dimensions other than `dim`
            return self.groupby(
                level=list(  # type: ignore
                    filter(lambda d: d not in dim, self.index.names)
                ),
                group_keys=False,
                observed=True,
            )

    def _replace(self, data) -> "AttrSeries":
        """Shorthand to preserve attrs."""
        return self.__class__(data, name=self.name, attrs=self.attrs)



from collections.abc import Generator, Sequence
from itertools import chain, tee
from operator import itemgetter
from typing import Any, Dict, Iterable, Optional, Union

from .key import Key, KeyLike


def _key_arg(key: KeyLike) -> Union[str, Key]:
    return Key.bare_name(key) or Key(key)


class Graph(dict):
    """A dictionary for a graph indexed by :class:`.Key`.

    Graph maintains indexes on set/delete/pop/update operations that allow for fast
    lookups/member checks in certain special cases:

    .. autosummary::

       unsorted_key
       full_key

    These basic features are used to provide higher-level helpers for
    :class:`.Computer`:

    .. autosummary::

       infer
    """

    _unsorted: Dict[KeyLike, KeyLike] = dict()
    _full: Dict[Key, Key] = dict()

    def __init__(self, *args, **kwargs):
        # Initialize members
        super().__init__(*args, **kwargs)

        # Initialize indices
        self._unsorted = dict()
        self._full = dict()

        # Index new keys
        for k in kwargs.keys():
            self._index(k)

    def _index(self, key: KeyLike):
        """Add `key` to the indices."""
        k = _key_arg(key)
        if isinstance(k, Key):
            self._unsorted[k.sorted] = k
            nodim = k.drop(True)
            if len(k.dims) >= len(self._full.get(nodim, nodim).dims):
                self._full[nodim] = k
        else:
            self._unsorted[k] = key

    def _deindex(self, key: KeyLike):
        """Remove `key` from the indices."""
        k = _key_arg(key)
        if isinstance(k, Key):
            self._unsorted.pop(k.sorted, None)
            self._full.pop(k.drop(True), None)
        else:
            self._unsorted.pop(k, None)

    def __setitem__(self, key: KeyLike, value: Any):
        super().__setitem__(key, value)
        self._index(key)

    def __delitem__(self, key: KeyLike):
        super().__delitem__(key)
        self._deindex(key)

    def __contains__(self, item) -> bool:
        """:obj:`True` if `item` *or* a key with the same dims in a different order."""
        try:
            return super().__contains__(item) or bool(self.unsorted_key(item))
        except Exception:  # for instance, TypeError
            return False

    def pop(self, *args):
        try:
            return super().pop(*args)
        finally:
            self._deindex(args[0])

    def update(self, arg=None, **kwargs):
        if isinstance(arg, (Sequence, Generator)):
            arg0, arg1 = tee(arg)
            arg_keys = map(itemgetter(0), arg0)
        else:
            arg1 = arg or dict()
            arg_keys = arg1.keys()

        for key in chain(kwargs.keys(), arg_keys):
            self._index(key)

        super().update(arg1, **kwargs)

    def unsorted_key(self, key: KeyLike) -> Optional[KeyLike]:
        """Return `key` with its original or unsorted dimensions."""
        k = _key_arg(key)
        return self._unsorted.get(k.sorted if isinstance(k, Key) else k)

    def full_key(self, name_or_key: KeyLike) -> Optional[KeyLike]:
        """Return `name_or_key` with its full dimensions."""
        return self._full.get(Key(name_or_key).drop_all())

    def infer(
        self, key: Union[str, Key], dims: Iterable[str] = []
    ) -> Optional[KeyLike]:
        """Infer a `key`.

        Parameters
        ----------
        dims : list of str, optional
            Drop all but these dimensions from the returned key(s).

        Returns
        -------
        str
            If `key` is not found in the Graph.
        Key
            `key` with either its full dimensions (cf. :meth:`full_key`) or, if `dims`
            are given, with only these dims.
        """
        result = self.unsorted_key(key) or key

        if isinstance(key, str) or not key.dims:
            # Find the full-dimensional key
            result = self.full_key(result) or ""

        if not isinstance(result, Key):
            return result or key

        # Drop all but `dims`
        if dims:
            result = result.drop(*(set(result.dims) - set(dims)))

        return result



import ast
import numbers
from functools import partial, singledispatchmethod
from textwrap import dedent

from dask.core import quote

from genno.core.key import Key
from genno.core.quantity import Quantity

BINOP = {
    ast.Add: "add",
    ast.Div: "div",
    ast.Mult: "mul",
    ast.Pow: "pow",
    ast.Sub: "sub",
}


def unparse(node) -> str:
    """Compatibility with Python ≤3.9, where :func:`ast.unparse` is not yet defined."""
    try:
        return ast.unparse(node)
    except AttributeError:  # Python 3.8
        return repr(node)


class Parser:
    """Parser for :meth:`.Computer.eval`."""

    def __init__(self, computer):
        self.computer = computer
        self.queue = []
        self.new_keys = {}

    def parse(self, expr: str):
        # - Remove leading/trailing newlines
        # - Dedent the entire string
        # - Parse the expression; returns ast.Module
        # - Iterate over the statements in the module body
        for statement in ast.parse(dedent(expr.strip("\n"))).body:
            self.recurse(statement)

    def append(self, operands, task, kwargs=None):
        # Construct the target key (maybe anonymous)
        key = Key.product(self.current_target, *operands)

        self.anonymous_tag += 1
        if self.anonymous_tag == 0:
            self.new_keys[key.name] = key
        else:
            key = Key(key.name, key.dims, tag=f"_{self.anonymous_tag}")

        # Add a task to the queue
        self.queue.append(((key,) + task, kwargs or dict()))

        # Return the constructed key to the outer call
        return key

    @singledispatchmethod
    def recurse(self, node):
        # Something else, probably an ast.BinOp
        try:
            return self.computer.get_comp(BINOP[node.__class__])
        except KeyError:
            raise NotImplementedError(f"ast.{node.__class__.__name__}")

    @recurse.register
    def _(self, node: ast.Assign):
        # Identify the target
        if len(node.targets) != 1:
            raise NotImplementedError(f"Assign to {len(node.targets)} != 1 targets")
        elif not isinstance(node.targets[0], ast.Name):
            raise NotImplementedError(f"Assign to {node.targets[0].__class__.__name__}")

        # Store the current target name
        self.current_target = node.targets[0].id

        # Reset counter
        self.anonymous_tag = -1

        # Iteratively parse the right-hand side
        self.recurse(node.value)

    @recurse.register
    def _(self, node: ast.BinOp):
        # A binary operation: recurse both sides and identify the operation
        op = self.recurse(node.op)
        left = self.recurse(node.left)
        right = self.recurse(node.right)

        # Return the constructed key to the outer call
        return self.append((left, right), (op, left, right))

    @recurse.register
    def _(self, node: ast.UnaryOp):
        # A unary operation: look up some portions of a task
        if isinstance(node.op, ast.USub):
            op = (self.computer.get_comp("mul"), Quantity(-1.0))
        else:
            raise NotImplementedError(f"ast.{node.op.__class__.__name__}")

        # Recurse the operand
        operand = self.recurse(node.operand)

        # Return the constructed key to the outer call
        return self.append((operand,), op + (operand,))

    @recurse.register
    def _(self, node: ast.Call):
        if not isinstance(node.func, ast.Name):
            raise NotImplementedError(
                f"Call {unparse(node.func)}(…) instead of function"
            )

        # Get the computation function
        func = self.computer.get_comp(node.func.id)
        if func is None:
            raise NameError(f"No computation named {node.func.id!r}")

        # Recurse args and keyword args
        args = [self.recurse(arg) for arg in node.args]
        keywords = dict(map(self.recurse, node.keywords))

        return self.append(tuple(args), tuple([partial(func, **keywords)] + args))

    @recurse.register
    def _(self, node: ast.Constant):
        # A constant: convert to a Quantity
        if isinstance(node.value, numbers.Real):
            return Quantity(node.value)
        else:
            # Something else, e.g. a string
            return quote(node.value)

    @recurse.register
    def _(self, node: ast.Name):
        # A name: either the name of an existing key, or one created in this eval()
        # call
        return self.new_keys.get(node.id, None) or self.computer.full_key(node.id)

    @recurse.register
    def _(self, node: ast.keyword):
        if not isinstance(node.value, ast.Constant):
            raise NotImplementedError(
                f"Non-literal keyword arg {unparse(node.value)!r}"
            )
        return (node.arg, node.value.value)



from functools import update_wrapper
from inspect import signature
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Dict, Optional, Tuple, Union

if TYPE_CHECKING:
    from .computer import Computer
    from .key import KeyLike


class Operator:
    """Base class for a callable with convenience methods.

    Example
    -------
    >>> from genno import Operator
    >>>
    >>> @Operator.define
    ... def myfunc(q1: Quantity, q2: Quantity) -> Quantity:
    ...     # Operator code
    >>>
    >>> @myfunc.helper
    ... def add_myfunc(f, computer, *args, **kwargs):
    ...     # Custom code to add tasks to `computer`
    ...     # Perform checks or handle `args` and `kwargs`.
    """

    # Use these specific attribute names to be intelligible to functools.partial()
    #: Function or callable for the Operator.
    func: ClassVar[Callable]
    args = ()
    keywords: Dict[str, Any] = dict()

    #: Helper method to add tasks to a :class:`.Computer`. Register with :meth:`helper`,
    #: invoke with :meth:`add_tasks`.
    _add_tasks: ClassVar[Optional[Callable]] = None

    def __call__(self, *args, **kwargs):
        # The callable is stored as a static method; `self` is not passed
        return self.func(*args, **kwargs)

    def __eq__(self, other):
        """Compares equal to the wrapped `func`."""
        return other == self.func

    def __repr__(self):
        return f"<operator {self.__class__.__name__}>"

    @staticmethod
    def define(func: Callable) -> "Operator":
        """Create an Operator object that wraps `func`."""
        # This follows the pattern of using a metaclass, except compressed

        # - Create the class
        #   - Same name as func, subclass of Operator
        #   - Subclass of Operator
        #   - func is a static method
        #   - Signature of klass.__call__ is the signature of func.
        klass = type(
            func.__name__,
            (Operator,),
            {"func": staticmethod(func), "__signature__": signature(func)},
        )

        # Create an instance of the class, update __doc__ and other attributes, return
        # NB these are updated on the instance, not on `klass`, to satisfy Sphinx, which
        #    will skip documenting items that have the same __doc__ as their class
        return update_wrapper(klass(), func, updated=())

    def helper(
        self, func: Callable[..., Union["KeyLike", Tuple["KeyLike", ...]]]
    ) -> Callable:
        """Register `func` as the convenience method for adding task(s)."""
        self.__class__._add_tasks = staticmethod(func)
        return func

    def add_tasks(self, c: "Computer", *args, **kwargs) -> Tuple["KeyLike", ...]:
        """Invoke :attr:`_add_task` to add tasks to `c`."""
        if self._add_tasks is None:
            raise NotImplementedError

        return self._add_tasks(self.func, c, *args, **kwargs)



import logging

import pytest

from genno import Computer, Quantity
from genno.testing import (
    assert_logs,
    assert_qty_allclose,
    assert_qty_equal,
    assert_units,
)

log = logging.getLogger(__name__)


@pytest.mark.xfail()
def test_assert_logs(caplog):
    caplog.set_level(logging.DEBUG)

    with assert_logs(caplog, "foo"):
        log.debug("bar")
        log.info("baz")
        log.warning("spam and eggs")


def test_assert_units():
    assert_units(Quantity(), "")


def test_assert_check_type():
    """Mismatched types in :func:`assert_qty_equal` and :func:`assert_qty_allclose`."""
    with pytest.raises(AssertionError):
        assert_qty_equal(int(1), 2.2)

    with pytest.raises(AssertionError):
        assert_qty_allclose(int(1), 2.2)


@pytest.mark.xfail(raises=TypeError)
def test_runtest_makereport():
    """The Pytest hook :func:`.pytest_runtest_makereport` works."""
    c = Computer()

    def func(x):
        return "a" + x

    c.add("test", func, 1.0)

    # This line raises a ComputationError, but the
    # genno.testing.pytest_runtest_makereport hook implementation unwraps the TypeError
    # that triggers it; this then satisfies the xfail marker on the test function.
    c.get("test")



import re
from operator import itemgetter

import pandas as pd
import pytest

from genno import Computer, Quantity
from genno.core.describe import label
from genno.core.sparsedataarray import SparseDataArray


@pytest.mark.usefixtures("parametrize_quantity_class")
def test_describe():
    c = Computer()
    c.add("foo", Quantity(pd.Series([42, 43])))

    if Quantity._get_class() is SparseDataArray:
        assert re.match(
            r"""'foo':
- <xarray\.SparseDataArray \([^:]+: 2\)>""",
            c.describe("foo"),
        )


def test_describe_shorten():
    c = Computer()
    c.add_single("foo", len, dict([(f"key{N}", "X" * N) for N in range(10)]))

    assert (
        """'config':
- {}

'foo':
- <built-in len>
- {'key0': '', 'key1': 'X', 'key2': 'XX', 'key3': 'XXX', 'key4': 'XXXX', 'key5': 'XXXXX', 'key6': 'XXXXXX', 'key7': 'XXXXXXX', 'key8': 'XXXXXXXX', 'key9': [...]

all"""  # noqa: E501
        == c.describe()
    )


def test_label():
    """:func:`label` handles unusual callables."""
    assert "operator.itemgetter(0)" == label(itemgetter(0))



try:
    import pyam  # noqa: F401
except ModuleNotFoundError:  # pragma: no cover
    HAS_PYAM = False
else:
    HAS_PYAM = True

import logging
from functools import partial
from typing import List

from genno import Computer, Key
from genno.config import handles
from genno.core.key import single_key

log = logging.getLogger(__name__)


@handles("iamc")
def iamc(c: Computer, info):
    """Handle one entry from the ``iamc:`` config section."""
    try:
        c.require_compat("pyam")
    except ModuleNotFoundError:  # pragma: no cover
        if not HAS_PYAM:
            log.warning("Missing pyam; configuration section 'iamc:' ignored")
            return
        else:
            raise

    from . import util

    # For each quantity, use a chain of computations to prepare it
    name = info.pop("variable")

    # Chain of keys produced: first entry is the key for the base quantity
    keys: List[Key] = [Key(info.pop("base"))]

    # Second entry is a simple rename
    keys.append(single_key(c.add_single(Key(name, keys[0].dims, keys[0].tag), keys[0])))

    # Optionally select a subset of data from the base quantity
    sel = info.get("select")
    if sel:
        keys.append(
            single_key(
                c.add_single(
                    keys[-1].add_tag("sel"),
                    (c.get_comp("select"), keys[-1], sel),
                    strict=True,
                )
            )
        )

    # Use a setting for the collapse callback function. This doesn't work from file,
    # since no way to define a Python function in JSON or YAML
    collapse_info = info.pop("collapse", {})
    collapse_func = collapse_info.pop("callback", util.collapse)

    # Use the Computer method to add the coversion step
    # NB convert_pyam() returns a single key when applied to a single key
    keys.append(
        single_key(
            c.add(
                keys[-1],
                "as_pyam",
                rename=info.pop("rename", {}),
                collapse=partial(collapse_func, **collapse_info),
                replace=info.pop("replace", {}),
                drop=set(info.pop("drop", [])) & set(keys[-1].dims),
                unit=info.pop("unit", None),
            )
        )
    )

    log.info(f"Add {repr(keys[-1])} from {repr(keys[0])}")
    log.debug(f"    {len(keys)} keys total")



import logging
from itertools import chain
from traceback import TracebackException, format_exception_only, format_list

log = logging.getLogger(__name__)


class ComputationError(Exception):
    """Wrapper to print intelligible exception information for :meth:`.Computer.get`.

    In order to aid in debugging, this helper:

    - Omits the parts of the stack trace that are internal to :mod:`dask`, and
    - Gives the key in the :attr:`.Computer.graph` and the computation/task that caused
      the exception.
    """

    def __init__(self, exc):
        self._exc = exc

    def __str__(self):
        """String representation.

        Most exception handling (Python, IPython, Jupyter) will print the traceback that
        led to `self` (i.e. the call to :meth:`.Computer.get`), followed by the string
        returned by this method.
        """
        try:
            return self._format()
        except Exception as format_exc:
            # Something went wrong during _format()
            log.error(
                f"Exception raised while formatting {self._exc}:\n" + repr(format_exc)
            )
            # Fall back to printing the underlying exception
            return str(self._exc)

    def _format(self):
        key, task, frames = process_dask_tb(self._exc)

        # Assemble the exception printout
        return "".join(
            chain(
                # Computer information for debugging
                [
                    f"computing {key} using:\n\n" if key else "",
                    f"{task}\n\n" if task else "",
                    "Use Computer.describe(...) to trace the computation.\n\n",
                    "Computation traceback:\n",
                ],
                # Traceback; omitting a few dask internal calls below execute_task
                format_list(frames),
                # Type and message of the original exception
                format_exception_only(self._exc.__class__, self._exc),
            )
        )


class KeyExistsError(KeyError):
    """Raised by :meth:`.Computer.add` when the target key exists."""

    def __str__(self):
        return f"key {repr(self.args[0])} already exists"


class MissingKeyError(KeyError):
    """Raised by :meth:`.Computer.add` when a required input key is missing."""

    def __str__(self):
        return f"required keys {repr(self.args)} not defined"


def process_dask_tb(exc):
    """Process *exc* arising from :meth:`.Computer.get`.

    Returns a tuple with 3 elements:

    - The key of the computation.
    - The info key of the computation.
    - A list of traceback.FrameSummary objects, without locals, for *only*
      frames that are not internal to dask.
    """
    key = task = None  # Info about the computation that triggered *exc*
    frames = []  # Frames for an abbreviated stacktrace

    try:
        # Get a traceback with captured locals
        tbe = TracebackException.from_exception(exc, capture_locals=True)
    except Exception:
        # Some exception occurred when capturing locals; proceed without
        tbe = TracebackException.from_exception(exc)

    # Iterate over frames from the base of the stack
    # Initial frames are internal to dask
    dask_internal = True
    for frame in tbe.stack:
        if frame.name == "execute_task":
            # Current frame is the dask internal call to execute a task
            try:
                # Retrieve information about the key/task that triggered the
                # exception. These are not the raw values of variables, but
                # their string repr().
                key = frame.locals["key"]
                task = frame.locals["task"]
            except (TypeError, KeyError):  # pragma: no cover
                # No locals, or 'key' or 'task' not present
                pass

            # Subsequent frames are related to the exception
            dask_internal = False

        if not dask_internal:
            # Don't display the locals when printing the traceback
            frame.locals = None

            # Store the frame for printing the traceback
            frames.append(frame)

    # Omit a few dask internal calls below execute_task
    return key, task, frames[3:]



"""Tests of AttrSeries in particular."""
import pandas as pd
import pandas.testing as pdt
import pytest

from genno import Computer
from genno.core.attrseries import AttrSeries
from genno.testing import add_large_data


class TestAttrSeries:
    @pytest.fixture
    def foo(self):
        idx = pd.MultiIndex.from_product([["a1", "a2"], ["b1", "b2"]], names=["a", "b"])
        yield AttrSeries([0, 1, 2, 3], index=idx, name="Foo", units="kg")

    @pytest.fixture
    def bar(self):
        """A 1-dimensional quantity."""
        yield AttrSeries([0, 1], index=pd.Index(["a1", "a2"], name="a"))

    def test_align_levels(self, foo, bar):
        # Scalar vs scalar
        q = AttrSeries(0.1)
        other = AttrSeries(2.2)
        _, result = q.align_levels(other)
        assert tuple() == result.dims

        # Scalar to 1D: broadcasts to 1D
        _, result = q.align_levels(bar)
        assert ("a",) == result.dims
        assert 1 == len(result.unique())

        # Scalar to 2D:
        _, result = q.align_levels(foo)
        assert ("b",) == result.dims
        assert 1 == len(result.unique())

        # 1D to scalar
        _, result = bar.align_levels(q)
        assert ("a",) == result.dims

        # 2D vs. 2D; same dimensions, different order
        idx = pd.MultiIndex.from_product([["b1", "b2"], ["a1", "a2"]], names=["b", "a"])
        q = AttrSeries([3, 2, 1, 0], index=idx)
        assert ("b", "a") == q.dims
        _, result = q.align_levels(foo)
        assert ("a", "b") == result.dims

    def test_cumprod(self, foo, bar):
        """AttrSeries.cumprod works with 1-dimensional quantities."""
        result0 = (1.1 + bar).cumprod("a")
        assert ("a",) == result0.dims

        # Same result with dim=None
        result1 = (1.1 + bar).cumprod()
        pdt.assert_series_equal(result0, result1)

        # But not with ≥1D
        with pytest.raises(NotImplementedError):
            foo.cumprod()

    def test_expand_dims(self, ureg, foo):
        # Name and units pass through expand_dims
        result = foo.expand_dims(c=["c1", "c2"])
        assert foo.name == result.name and ureg.Unit("kg") == foo.units == result.units

    def test_interp(self, foo):
        with pytest.raises(NotImplementedError):
            foo.interp(coords=dict(a=["a1", "a1.5", "a2"], b=["b1", "b1.5", "b2"]))

    def test_rename(self, foo):
        assert foo.rename({"a": "c", "b": "d"}).dims == ("c", "d")

    @pytest.mark.parametrize("indexers_kwargs", (dict(a="a2"), dict(a=["a2"])))
    def test_sel(self, bar, indexers_kwargs) -> None:
        # Selecting 1 element from 1-D parameter still returns AttrSeries
        result = bar.sel(**indexers_kwargs)
        assert isinstance(result, AttrSeries)
        assert result.size == 1
        assert result.dims == ("a",)
        assert result.iloc[0] == 1

    def test_sel_not_implemented(self, bar):
        with pytest.raises(NotImplementedError):
            bar.sel(a="a2", method="bfill")

        with pytest.raises(NotImplementedError):
            bar.sel(a="a2", tolerance=0.01)

    def test_shift(self, foo):
        foo.shift(a=1)
        foo.shift(b=1)

    def test_squeeze(self, foo):
        assert foo.sel(a="a1").squeeze().dims == ("b",)
        assert foo.sel(a="a2", b="b1").squeeze().values == 2

        with pytest.raises(
            ValueError,
            match="dimension to squeeze out which has length greater than one",
        ):
            foo.squeeze(dim="b")

        with pytest.raises(KeyError, match="c"):
            foo.squeeze(dim="c")

    def test_sum(self, foo, bar):
        # AttrSeries can be summed across all dimensions
        result = foo.sum(dim=["a", "b"])
        assert isinstance(result, AttrSeries)  # returns an AttrSeries
        assert result.size == 1  # with one element
        assert result.item() == 6  # that has the correct value

        # Sum with wrong dim raises ValueError
        with pytest.raises(ValueError):
            bar.sum("b")

        # Index with duplicate entries
        _baz = pd.DataFrame(
            [
                ["a1", "b1", "c1", 1.0],
                ["a1", "b1", "c1", 2.0],
                ["a2", "b2", "c3", 3.0],
                ["a2", "b2", "c4", 4.0],
                ["a3", "b3", "c5", 5.0],
            ],
            columns=["a", "b", "c", "value"],
        )
        # Fails with v1.13.0 AttrSeries.sum() using unstack()
        AttrSeries(_baz.set_index(["a", "b", "c"])["value"]).sum(dim="c")

        with pytest.raises(NotImplementedError):
            bar.sum("a", skipna=False)

    def test_others(self, foo, bar):
        # Exercise other compatibility functions
        assert type(foo.to_frame()) is pd.DataFrame
        assert foo.drop("a").dims == ("b",)
        assert bar.dims == ("a",)

        with pytest.raises(NotImplementedError):
            bar.item("a2")
        with pytest.raises(ValueError):
            bar.item()


@pytest.mark.skip(reason="Slow, for benchmarking only")
def test_sum_large(N_data=1e7):  # pragma: no cover
    """Test :meth:`.AttrSeries.sum` for large, sparse data."""
    # Create a single large AttrSeries
    c = Computer()
    add_large_data(c, 1, N_dims=11, N_data=N_data)
    qty = c.get("q_00")

    # Compute a sum()
    result = qty.sum(dim=["j", "k"])
    # print(result)  # DEBUG
    del result



import logging
import random
import re
from contextlib import nullcontext
from functools import partial
from typing import Hashable, Iterable, Mapping

import numpy as np
import pandas as pd
import pint
import pytest
import xarray as xr
from dask.core import quote
from numpy.testing import assert_allclose
from pandas.testing import assert_series_equal

from genno import Computer, Quantity, computations
from genno.core.sparsedataarray import SparseDataArray
from genno.testing import (
    add_large_data,
    add_test_data,
    assert_logs,
    assert_qty_allclose,
    assert_qty_equal,
    random_qty,
)

pytestmark = pytest.mark.usefixtures("parametrize_quantity_class")


@pytest.fixture(scope="function")
def data():
    """Yields a computer, then the values of :func:`.add_test_data`."""
    c = Computer()
    yield [c] + list(add_test_data(c))


@pytest.mark.parametrize(
    "operands, size",
    [
        (("a", "a"), 18),
        (("a", "x"), 36),
        (("x", "b"), 36),
        (("a", "b"), 36),
        (("a", "x", "b"), 36),
    ],
)
def test_add(data, operands, size):
    # Unpack
    c, t, t_foo, t_bar, x = data

    y = c.get("y")
    x = c.get("x:t-y")
    a = Quantity(
        xr.DataArray(
            np.random.rand(len(t_foo), len(y)), coords=[t_foo, y], dims=["t", "y"]
        ),
        units=x.units,
    )
    b = Quantity(
        xr.DataArray(
            np.random.rand(len(t_bar), len(y)), coords=[t_bar, y], dims=["t", "y"]
        ),
        units=x.units,
    )

    c.add("a:t-y", a)
    c.add("b:t-y", b)

    key = c.add(
        "result", tuple([computations.add] + [f"{name}:t-y" for name in operands])
    )

    result = c.get(key)
    assert size == result.size, result.to_series()


def test_add_units():
    """Units are handled correctly by :func:`.add`."""
    A = Quantity(1.0, units="kg")
    B = Quantity(1.0, units="tonne")

    # Units of result are units of the first argument
    assert_qty_equal(Quantity(1001.0, units="kg"), computations.add(A, B))
    assert_qty_equal(Quantity(1.001, units="tonne"), computations.add(B, A))

    with pytest.raises(ValueError, match="Units 'kg' and 'km' are incompatible"):
        computations.add(A, Quantity(1.0, units="km"))


@pytest.mark.parametrize("keep", (True, False))
def test_aggregate0(caplog, data, keep):
    *_, t_foo, t_bar, x = data

    t_groups = dict(foo=t_foo, bar=t_bar)

    result = computations.aggregate(x, dict(t=t_groups), keep)
    assert result.name == x.name and result.units == x.units  # Pass through

    # Result has the expected dimensions
    assert set(t_groups) | (set(t_foo + t_bar) if keep else set()) == set(
        result.coords["t"].data
    )

    # Now with a group ID that duplicates one of the existing index names
    t_groups[t_foo[0]] = t_foo[:1]
    with (
        assert_logs(
            caplog, f"t='{t_foo[0]}' is already present in quantity 'Quantity X'"
        )
        if keep
        else nullcontext()
    ):
        result = computations.aggregate(x, dict(t=t_groups), keep)

    # Two dimensions
    result = computations.aggregate(x, {"t": t_groups, "y": {"2k": [2000, 2010]}}, keep)
    assert "2k" in result.coords["y"]


@pytest.mark.parametrize("keep", (True, False))
def test_aggregate_regex(caplog, data, keep):
    """:func:`aggregate` using regular expressions."""
    *_, t_foo, t_bar, x = data

    # Aggregate using regular expressions
    t_groups = dict(foo=[re.compile("foo[123]")], bar=[re.compile("b[ar]{2}.")])

    # Operation completes
    result = computations.aggregate(x, dict(t=t_groups), keep)

    # Name and units pass through
    assert result.name == x.name and result.units == x.units

    # Result has the expected dimensions
    assert set(t_groups) | (set(t_foo + t_bar) if keep else set()) == set(
        result.coords["t"].data
    )


def test_apply_units(data, caplog):
    # Unpack
    *_, x = data

    registry = pint.get_application_registry()

    # Brute-force replacement with incompatible units
    with assert_logs(
        caplog, "Replace 'kilogram' with incompatible 'liter'", at_level=logging.DEBUG
    ):
        result = computations.apply_units(x, "litres")
    assert result.units == registry.Unit("litre")
    # No change in values
    assert_series_equal(result.to_series(), x.to_series())
    assert result.name == x.name  # Pass through

    # Compatible units: magnitudes are also converted
    with assert_logs(
        caplog, "Convert 'kilogram' to 'metric_ton'", at_level=logging.DEBUG
    ):
        result = computations.apply_units(x, "tonne")
    assert result.units == registry.Unit("tonne")
    assert_series_equal(result.to_series(), x.to_series() * 0.001)

    # Remove unit
    x.units = registry.Unit("dimensionless")

    caplog.clear()
    result = computations.apply_units(x, "kg")
    # Nothing logged when _unit attr is missing
    assert len(caplog.messages) == 0
    assert result.units == registry.Unit("kg")
    assert_series_equal(result.to_series(), x.to_series())


def test_assign_units(data, caplog):
    # Unpack
    *_, x = data

    registry = pint.get_application_registry()

    # Brute-force replacement with incompatible units
    with assert_logs(
        caplog,
        "Replace 'kilogram' with 'liter' with different dimensionality",
        at_level=logging.INFO,
    ):
        result = computations.assign_units(x, "litres")
    assert result.units == registry.Unit("litre")
    # No change in values
    assert_series_equal(result.to_series(), x.to_series())
    assert result.name == x.name  # Pass through

    # Compatible units: magnitudes are not changed
    with assert_logs(
        caplog,
        "Replace 'kilogram' with 'metric_ton' without altering magnitudes",
        at_level=logging.INFO,
    ):
        result = computations.assign_units(x, "tonne")
    assert result.units == registry.Unit("tonne")
    assert_series_equal(result.to_series(), x.to_series())

    # Remove unit
    x.units = registry.Unit("dimensionless")

    caplog.clear()
    result = computations.assign_units(x, "kg")
    # Nothing logged when _unit attr is missing
    assert len(caplog.messages) == 0
    assert result.units == registry.Unit("kg")
    assert_series_equal(result.to_series(), x.to_series())


def test_convert_units(data, caplog):
    # Unpack
    *_, x = data

    registry = pint.get_application_registry()

    # Brute-force replacement with incompatible units
    with pytest.raises(ValueError, match="cannot be converted to"):
        result = computations.convert_units(x, "litres")

    # Compatible units: magnitudes are also converted
    result = computations.convert_units(x, "tonne")
    assert registry.Unit("tonne") == result.units
    assert_series_equal(
        result.to_series(), (x.to_series() * 0.001).rename("Quantity X")
    )
    assert result.name == x.name  # Pass through

    # Remove unit
    x.units = registry.Unit("dimensionless")

    with pytest.raises(ValueError, match="cannot be converted to"):
        result = computations.convert_units(x, "kg")


@pytest.mark.parametrize(
    "map_values, kwarg",
    (
        ([[1.0, 1, 0], [0, 0, 1]], dict()),
        pytest.param(
            [[1.0, 1, 0], [0, 1, 1]],
            dict(strict=True),
            marks=pytest.mark.xfail(raises=ValueError, reason="invalid map"),
        ),
    ),
)
def test_broadcast_map(ureg, map_values, kwarg):
    x = ["x1"]
    y = ["y1", "y2"]
    z = ["z1", "z2", "z3"]
    q = Quantity(xr.DataArray([[42.0, 43]], coords=[("x", x), ("y", y)]))
    m = Quantity(xr.DataArray(map_values, coords=[("y", y), ("z", z)]))

    result = computations.broadcast_map(q, m, **kwarg)
    exp = Quantity(
        xr.DataArray([[42.0, 42, 43]], coords=[("x", x), ("z", z)]),
        units=ureg.dimensionless,
    )

    assert_qty_equal(exp, result)


def test_combine(ureg, data):
    *_, t_bar, x = data

    # Without select, preserves the "t" dimension
    result = computations.combine(x, x, x, weights=(-1, 0.2, 0.8))

    assert ("t", "y") == result.dims
    assert 36 == result.size
    assert all(1e-15 > result.to_series().values)

    # With select, the selected values are summed along the "t" dimension
    result = computations.combine(
        x, x, select=(dict(t=t_bar), dict(t=t_bar)), weights=(-1, 1)
    )

    assert ("y",) == result.dims
    assert 6 == result.size
    assert all(1e-15 > result.to_series().values)

    # Incompatible units raises ValueError
    x2 = Quantity(x, units=ureg.metre)
    with pytest.raises(
        ValueError, match=re.escape("Cannot combine() units kilogram and meter")
    ):
        computations.combine(
            x, x2, select=(dict(t=t_bar), dict(t=t_bar)), weights=(-1, 1)
        )


def test_concat(ureg, data):
    *_, t_foo, t_bar, x = data

    # Split x into two concatenateable quantities
    a = computations.select(x, dict(t=t_foo))
    b = computations.select(x, dict(t=t_bar))

    # Concatenate
    computations.concat(a, b, dim="t")

    # Concatenate twice on a new dimension
    result = computations.concat(x, x, dim=pd.Index(["z1", "z2"], name="z"))

    # NB for AttrSeries, the new dimension is first; for SparseDataArray, last
    assert {"t", "y", "z"} == set(result.dims)
    assert ureg.Unit("kg") == x.units == result.units


def test_concat_dim_order(data):
    """:func:`.concat` succeeds even if dimension are not in matching order on operands.

    Test of https://github.com/khaeru/genno/issues/38.
    """
    *_, x = data

    # Create another Quantity like `x`, but with dims in the opposite order
    z = Quantity(
        x.to_series()
        .reset_index()
        .eval("y = y + 1000")
        .set_index(list(reversed(x.dims)))
        .iloc[:, 0]
    )
    assert tuple(reversed(x.dims)) == z.dims

    # Concatenation succeeds
    result = computations.concat(x, z)

    # Dims and length are as expected
    assert set(x.dims) == set(z.dims) == set(result.dims)
    assert len(x) + len(z) == len(result)

    # Dimensions were aligned correctly such that "y" in the result contains labels from
    # the same dimension in both operands
    assert set(x.coords["y"].data) | set(z.coords["y"].data) == set(
        result.coords["y"].data
    )


@pytest.mark.parametrize("func", [computations.div, computations.ratio])
def test_div(func, ureg):
    # Non-overlapping dimensions can be broadcast together
    A = random_qty(dict(x=3, y=4), units="km")
    B = random_qty(dict(z=2), units="hour")

    result = func(A, B)
    assert ("x", "y", "z") == result.dims
    assert ureg.Unit("km / hour") == result.units


def test_div_scalar(data, ureg):
    """:func:`.div` handles scalar numerator."""

    *_, x = data

    result = computations.div(2.0, x)
    assert ("t", "y") == result.dims
    assert_qty_equal(
        2.0 / x.sel(t="foo1", y=2000), result.sel(t="foo1", y=2000), check_attrs=False
    )


def test_drop_vars(data):
    # Unpack
    *_, x = data

    result = computations.drop_vars(x, "t")
    assert result.name == x.name and result.units == x.units  # Pass through
    assert set(x.dims) == {"t"} | set(result.dims)


def test_group_sum(ureg):
    a = "a1 a2".split()
    b = "b1 b2 b3".split()
    X = Quantity(
        xr.DataArray(np.random.rand(2, 3), coords=[("a", a), ("b", b)]),
        units=ureg.kg,
        name="Foo",
    )

    result = computations.group_sum(X, "a", "b")
    assert result.name == X.name and result.units == X.units  # Pass through
    assert ("a",) == result.dims
    assert 2 == len(result)


def test_index_to(ureg):
    q = random_qty(dict(x=3, y=5))
    q.name = "Foo"
    q.units = ureg.kg

    exp = q / q.sel(x="x0")
    exp.units = ""  # Indexed values are dimensionless

    # Called with a mapping
    result = computations.index_to(q, dict(x="x0"))
    assert_qty_equal(exp, result)
    assert exp.name == result.name

    # Called with two positional arguments
    result = computations.index_to(q, "x", "x0")
    assert_qty_equal(exp, result)

    # Default first index selected if 'None' is given
    result = computations.index_to(q, "x")
    assert_qty_equal(exp, result)

    result = computations.index_to(q, dict(x=None))
    assert_qty_equal(exp, result)

    # Invalid calls
    with pytest.raises(TypeError, match="expected a mapping from 1 key to 1 value"):
        computations.index_to(q, dict(x="x0", y="y0"))  # Length != 1

    with pytest.raises(KeyError):
        computations.index_to(q, dict(x="x99"))  # Mapping to something invalid


@pytest.mark.parametrize(
    "shape",
    [
        dict(x=3),
        dict(x=3, y=3),
        dict(y=3, x=3),
        dict(x=3, y=3, z=2),
        dict(y=3, x=3, z=2),
        dict(y=3, z=2, x=3),
    ],
)
def test_interpolate(caplog, ureg, shape):
    """Test :func:`.interpolate`."""
    # Generate a random quantity with one dimension indexed by integers
    q = random_qty(shape, name="Foo", units=ureg.kg)
    x = [2020, 2030, 2040]
    q = q.assign_coords({"x": x})

    # Linear interpolation of 1 point
    result = computations.interpolate(q, dict(x=2025), assume_sorted=False)
    assert result.name == q.name and result.units == q.units  # Pass through
    assert "interpolate(…, assume_sorted=False) ignored" in caplog.messages

    # Result has the expected class, dimensions, and values
    assert isinstance(result, q.__class__)
    assert tuple([d for d in q.dims if d != "x"]) == result.dims
    assert_qty_allclose(
        result, 0.5 * q.sel(x=[2020, 2030]).sum("x"), ignore_extra_coords=True
    )

    # Extrapolation on both ends of the data
    x = sorted(x + [x[0] - 1, x[-1] + 1])

    # interpolate() works
    result = computations.interpolate(
        q, dict(x=x), method="linear", kwargs=dict(fill_value="extrapolate")
    )

    # Produces the expected results
    r = result
    for i1, i2, i3 in ((0, 1, 2), (-1, -2, -3)):
        # Slope interior to the existing data
        slope_int = (r.sel(x=x[i3], drop=True) - r.sel(x=x[i2], drop=True)) / (
            x[i3] - x[i2]
        )
        # Slope to extrapolated points
        slope_ext = (r.sel(x=x[i2], drop=True) - r.sel(x=x[i1], drop=True)) / (
            x[i2] - x[i1]
        )
        # print(
        #     (i1, x[i1], r.sel(x=x[i1])),
        #     (i2, x[i2], r.sel(x=x[i2])),
        #     (i3, x[i3], r.sel(x=x[i3])),
        #     slope_int,
        #     slope_ext,
        # )
        assert_qty_allclose(slope_int, slope_ext)


@pytest.mark.parametrize(
    "name, kwargs",
    [
        ("input0.csv", dict(units="km")),
        # Units kwarg as a pint.Quantity
        ("input0.csv", dict(units=pint.get_application_registry()("1.0 km"))),
        # Dimensions as a container, without mapping
        ("input0.csv", dict(dims=["i", "j"], units="km")),
        #
        # Map a dimension name from the file to a different one in the quantity; ignore
        # dimension "foo"
        ("input1.csv", dict(dims=dict(i="i", j_dim="j"))),
        ("input2.csv", dict(dims=["i", "j"])),
        ("input2.csv", dict(dims=["i", "j"], units="km")),  # Logs a warning
        # Exceptions
        pytest.param(
            "input1.csv",
            dict(dims=dict(i="i", j_dim="j"), units="kg"),
            marks=pytest.mark.xfail(
                raises=ValueError, reason="Explicit units 'kg' do not match 'km'…"
            ),
        ),
        pytest.param(
            "load_file-invalid.csv",
            dict(),
            marks=pytest.mark.xfail(
                raises=ValueError, reason="with non-unique units array(['cm'], ['km'],"
            ),
        ),
    ],
)
def test_load_file(test_data_path, ureg, name, kwargs):
    qty = computations.load_file(test_data_path / name, name="baz", **kwargs)

    assert ("i", "j") == qty.dims
    assert ureg.kilometre == qty.units
    assert "baz" == qty.name


@pytest.mark.parametrize(
    "func",
    (
        computations.mul,
        computations.product,  # Alias
        computations.disaggregate_shares,  # Deprecated alias
    ),
)
def test_mul0(func):
    A = Quantity(xr.DataArray([1.0, 2], coords=[("a", ["a0", "a1"])]))
    B = Quantity(xr.DataArray([3.0, 4], coords=[("b", ["b0", "b1"])]))
    exp = Quantity(
        xr.DataArray(
            [[3.0, 4], [6, 8]],
            coords=[("a", ["a0", "a1"]), ("b", ["b0", "b1"])],
        ),
        units="1",
    )

    assert_qty_equal(exp, func(A, B))


@pytest.mark.parametrize("func", [computations.mul, computations.product])
@pytest.mark.parametrize(
    "dims, exp_dims, exp_shape",
    (
        # Scalar × scalar
        (({}, {}), tuple(), tuple()),
        # Scalar × 1D
        (({}, dict(a=2)), ("a",), (2,)),
        # 1D × scalar
        ((dict(a=2), {}), ("a",), (2,)),
        # 1D with disjoint dimensions ** 3 = 3D
        ((dict(a=2), dict(b=2), dict(c=2)), tuple("abc"), (2, 2, 2)),
        # 2D × scalar × scalar = 2D
        ((dict(a=2, b=2), {}, {}), tuple("ab"), (2, 2)),
        # scalar × 1D × scalar = 1D
        (({}, dict(a=2), {}), tuple("a"), (2,)),
        # 4D × 5D, with some overlapping dimensions
        (
            (dict(a=2, b=2, c=2, d=2), dict(b=2, c=2, d=2, e=2, f=2)),
            tuple("abcdef"),
            (2, 2, 2, 2, 2, 2),
        ),
    ),
)
def test_mul1(func, dims, exp_dims, exp_shape):
    """Product of quantities with disjoint and overlapping dimensions."""
    quantities = [random_qty(d) for d in dims]

    result = func(*quantities)

    assert exp_dims == result.dims
    assert exp_shape == result.shape


@pytest.mark.parametrize(
    "exponent, base_units, exp_units",
    (
        # 2D dimensionless ** float
        (1.2, None, None),
        # 2D with units ** int
        (2, "kg", "kg ** 2"),
        # 2D ** 1D int
        (pd.Series(dict(y1=1, y2=2, y3=3)), "kg", ""),
        # 2D ** 1D int, all values the same
        (pd.Series(dict(y1=2, y2=2, y3=2)), "kg", "kg ** 2"),
        # 2D ** 1D with units
        pytest.param(
            random_qty(dict(y=3), units="km"),
            None,
            None,
            marks=pytest.mark.xfail(
                raises=ValueError, reason="Cannot raise to a power with units (km)"
            ),
        ),
    ),
)
def test_pow_simple(ureg, exponent, base_units, exp_units):
    A = random_qty(dict(x=3, y=3), units=base_units)

    # Convert using the current Quantity class
    if isinstance(exponent, pd.Series):
        exponent = Quantity(exponent)

    result = computations.pow(A, exponent)
    assert exp_units is None or ureg.Unit(exp_units) == result.units


def test_pow(ureg):
    A = random_qty(dict(x=3, y=3))

    # 2D dimensionless ** int
    result = computations.pow(A, 2)
    # Expected values
    assert_qty_equal(A.sel(x="x1", y="y1") ** 2, result.sel(x="x1", y="y1"))

    # 2D ** 1D float
    B = random_qty(dict(y=3))
    result = computations.pow(A, B)
    # Expected values
    assert_allclose(
        A.sel(x="x1", y="y1").item() ** B.sel(y="y1").item(),
        result.sel(x="x1", y="y1").item(),
    )
    assert ureg.dimensionless == result.units


def test_relabel(data):
    # Unpack
    c, t, t_foo, t_bar, x = data

    # Mapping from old to new labels for each dimension
    args = dict(
        t={"foo2": "baz", "bar5": "qux", "nothing": "nothing"},
        y={2030: 3030},
        not_a_dimension=None,
    )

    def check(qty):
        # Dimension t was relabeled
        t_out = set(qty.coords["t"].data)
        assert {"baz", "qux"} < t_out and not {"foo2", "bar5"} & t_out
        # Dimension y was relabeled
        assert 3030 in result.coords["y"] and 2030 not in result.coords["y"]

    # Can be called with a dictionary
    result = computations.relabel(x, args)
    check(result)
    assert result.name == x.name and result.units == x.units  # Pass through

    # Can be added and used through Computer

    # Store the name map in the Computer
    c.add("labels", quote(args))

    # Test multiple ways of adding this computation
    for args in [
        ("test", computations.relabel, "x:t-y", args),
        ("test", partial(computations.relabel, **args), "x:t-y"),
        ("test", "relabel", "x:t-y", args),
        ("test", "relabel", "x:t-y", "labels"),
        # Deprecated
        ("relabel", "test", "x:t-y", args),
        ("relabel", "test", "x:t-y", "labels"),
    ]:
        c.add(*args)
        result = c.get("test")
        check(result)


def test_rename_dims(data):
    # Unpack
    c, t, t_foo, t_bar, x = data

    # Can be called with a dictionary
    args = {"t": "s", "y": "z"}
    result = computations.rename_dims(x, args)
    assert result.name == x.name and result.units == x.units  # Pass through
    assert ("s", "z") == result.dims  # Quantity has renamed dimensions
    assert all(t == result.coords["s"])  # Renamed dimension contain original labels

    # Can be called with keyword arguments
    result = computations.rename_dims(x, **args)
    assert ("s", "z") == result.dims and all(t == result.coords["s"])  # As above

    with pytest.raises(ValueError, match="cannot specify both keyword and positional"):
        computations.rename_dims(x, args, **args)

    # Can be added and used through Computer

    # Store the name map in the Computer
    c.add("dim name map", quote(args))

    # Test multiple ways of adding this computation
    for args in [
        ("test", computations.rename_dims, "x:t-y", args),
        ("test", partial(computations.rename_dims, **args), "x:t-y"),
        ("test", "rename_dims", "x:t-y", args),
        ("test", "rename_dims", "x:t-y", "dim name map"),
        # Deprecated
        ("rename_dims", "test", "x:t-y", args),
        ("rename_dims", "test", "x:t-y", "dim name map"),
    ]:
        c.add(*args)
        result = c.get("test")
        assert ("s", "z") == result.dims and all(t == result.coords["s"])  # As above


def test_round(data):
    # Unpack
    *_, x = data

    # Up to 36 unique random values in `x`
    assert 2 < len(x.to_series().unique()) <= 36

    # round() runs
    result0 = computations.round(x)
    assert result0.name == x.name and result0.units == x.units

    # Only 0 or 1
    assert {0.0, 1.0} >= set(result0.to_series().unique())

    # round to 1 decimal place
    result1 = computations.round(x, 1)
    assert 0 <= len(result1.to_series().unique()) <= 11


@pytest.mark.parametrize(
    "indexers, kwargs, exp_dims, exp_shape",
    (
        # Length-1 indexer but drop=False (default) results in 2D data
        ({"y": [2010]}, {}, ("t", "y"), (6, 1)),
        # Same, with an additional non-existent label → same result
        ({"y": [2010, 9999]}, {}, ("t", "y"), (6, 1)),
        # Scalar indexer with drop=True results in 1D data
        ({"y": 2010}, dict(drop=True), ("t",), (6,)),
        # Length-1 indexer with drop=True results in 2D data
        ({"y": [2010]}, dict(drop=True), ("t", "y"), (6, 1)),
        # Scalar indexer with nonexistent label -> KeyError
        pytest.param({"y": 9999}, {}, (), (), marks=pytest.mark.xfail(raises=KeyError)),
        # Length-1 indexer with nonexistent label
        # NB this gives shape (6, 0) with SparseDataArray; (0, 0) with AttrSeries; test
        #    the size instead
        ({"y": [9999]}, {}, ("t", "y"), 0),
        # With inverse=True, the given label is dropped from the given dimension
        ({"y": [2010]}, dict(inverse=True), ("t", "y"), (6, 6 - 1)),
    ),
)
def test_select0(data, indexers, kwargs, exp_dims, exp_shape) -> None:
    *_, x = data

    result = computations.select(x, indexers=indexers, **kwargs)

    assert exp_dims == result.dims
    if isinstance(exp_shape, tuple):
        assert exp_shape == result.shape
    else:
        assert exp_shape == result.size


def test_select1(data) -> None:
    # Unpack
    *_, t_foo, t_bar, x = data

    N_y = 6

    x = Quantity(x)
    assert x.size == 6 * N_y

    # Selection with inverse=False
    indexers: Mapping[Hashable, Iterable[Hashable]] = {"t": t_foo[0:1] + t_bar[0:1]}
    result_0 = computations.select(x, indexers=indexers)
    assert result_0.size == 2 * N_y
    assert result_0.name == x.name and result_0.units == x.units  # Pass through

    # Select with labels that do not appear in the data
    result_1 = computations.select(x, indexers={"t": t_foo + ["MISSING"]})
    assert result_1.size == len(t_foo) * N_y

    # Select with xarray indexers
    indexers = {
        "t": xr.DataArray(
            t_foo, dims="new_dim", coords={"new_dim": ["d1", "d2", "d3"]}
        ),
        "y": xr.DataArray(
            [2000, 2010, 2020], dims="new_dim", coords={"new_dim": ["d1", "d2", "d3"]}
        ),
    }
    # NB with pandas 2.1, this triggers the RecursionError fixed in khaeru/genno#99
    result_2 = computations.select(x, indexers)
    assert ("new_dim",) == result_2.dims

    with pytest.raises(NotImplementedError):
        computations.select(x, indexers, inverse=True)


def test_select_bigmem():
    """:func:`select` with large quantities does not exhaust memory."""
    # Create a Computer with large data
    c = Computer()
    keys = add_large_data(c, num_params=2, N_dims=17)

    # Add a task top generate random indexers
    def random_indexers(qty, *, dim_index=0, k=20):
        dims = qty.dims
        coords = qty.coords

        d = dims[dim_index]
        return {d: list(map(lambda c: c.item(), random.choices(coords[d], k=k)))}

    k = c.add("random indexers", random_indexers, keys[0])

    # Add a task to select some values
    key = c.add("test key", "select", keys[0], k)

    if Quantity._get_class() is SparseDataArray:
        # ValueError: invalid dims: array size defined by dims is larger than the
        # maximum possible size.
        pytest.xfail(reason="Too large for sparse")

    # Selection occurs without raising MemoryError or segfault
    result = c.get(key)

    # Result can be converted to pd.Series
    result.to_series()


def test_sub(data):
    *_, t_foo, t_bar, x = data

    # Function runs
    result = computations.sub(x, x)

    assert (0 == result).all()

    assert result.units == x.units  # Pass through


@pytest.mark.parametrize("dimensions", (["t"], ["y"], ["t", "y"]))
def test_sum(data, dimensions):
    *_, t_foo, t_bar, x = data

    # Function runs
    result = computations.sum(x, dimensions=dimensions)

    assert result.name == x.name and result.units == x.units  # Pass through



import json
import logging
import pickle
from functools import partial, singledispatch, update_wrapper
from hashlib import blake2b
from inspect import getmembers, iscode
from pathlib import Path
from typing import Callable, Union

from .util import unquote

log = logging.getLogger(__name__)

# Types to ignore in Encoder.default()
IGNORE = set()


@singledispatch
def _encode(o):
    # Let the base class default method raise the TypeError
    return json.JSONEncoder().default(o)


@_encode.register
def _encode_path(o: Path):
    return str(o)


class Encoder(json.JSONEncoder):
    """JSON encoder.

    This is a one-way encoder used only to serialize arguments for :func:`.hash_args`
    and :func:`.hash_code`.
    """

    @classmethod
    def ignore(cls, *types):
        """Tell the Encoder (thus :func:`.hash_args`) to ignore arguments of `types`.

        Example
        -------
        >>> class Bar:
        >>>    pass
        >>>
        >>> # Don't use Bar instances in cache keys
        >>> @genno.caching.Encoder.ignore(Bar)

        Ignore all unrecognized types

        >>> @genno.caching.Encoder.ignore(object)
        """
        IGNORE.add(types)

    @classmethod
    def register(cls, func):
        """Register a `func` to serialize a type not handled by :class:`.JSONEncoder`.

        `func` should return a type that *is* handled by JSONEncoder; see the docs.

        Example
        -------
        >>> class Foo:
        >>>    a = 3
        >>>
        >>> @genno.caching.Encoder.register
        >>> def _encode_foo(o: Foo):
        >>>     return dict(a=o.a)  # JSONEncoder can handle dict()
        """
        return _encode.register(func)

    def default(self, o):
        """For `o`, return an object serializable by the base :class:`.JSONEncoder`.

        - :class:`pathlib.Path`: the string representation of `o`.
        - :class:`code` objects (from Python's built-in :mod:`inspect` module), e.g.
          a function or lambda: :class:`blake2b` hash of the object's bytecode and its
          serialized constants.

          .. warning:: This is not 100% guaranteed to change if the operation of `o` (or
             other code called in turn by `o`) changes. If relying on this behaviour,
             check carefully.
        - Any type indicated with :meth:`.ignore`: empty :class:`tuple`.
        - Any type with a serializer registered with :meth:`.register`: the return value
          of the serializer, called on `o`.
        """

        if iscode(o):
            # Python built-in code object, from e.g. hash_code() or lambda: hash the
            # identifying information: raw bytecode & constants used
            return blake2b(
                o.co_code + json.dumps(o.co_consts, cls=self.__class__).encode()
            ).hexdigest()

        try:
            return _encode(o)
        except TypeError:
            if isinstance(o, tuple(IGNORE)):
                log.warning(f"Cache key ignores {type(o)}")
                return ()
            else:
                raise


def hash_args(*args, **kwargs):
    """Return a 20-character :class:`blake2b` hex digest of `args` and `kwargs`.

    Used by :func:`.decorate`.

    See also
    --------
    Encoder
    """
    return blake2b(
        (
            ""
            if len(args) + len(kwargs) == 0
            else json.dumps((args, kwargs), cls=Encoder)
        ).encode(),
        digest_size=20,
    ).hexdigest()


def hash_code(func: Callable) -> str:
    """Return the :class:`.blake2b` hex digest of the compiled bytecode of `func`.

    See also
    --------
    Encoder
    """
    # Get the code object
    code_obj = next(filter(lambda kv: kv[0] == "__code__", getmembers(func)))[1]
    return Encoder().default(code_obj)


def hash_contents(path: Union[Path, str], chunk_size=65536) -> str:
    """Return the :class:`.blake2b` hex digest of the contents of the file at `path`.

    Parameters
    ----------
    chunk_size : int, optional
        Read the file in chunks of this size; default 64 kB.
    """
    with Path(path).open("rb") as f:
        hash = blake2b()
        for chunk in iter(partial(f.read, chunk_size), b""):
            hash.update(chunk)
    return hash.hexdigest()


def decorate(
    func: Callable, computer=None, cache_path=None, cache_skip=False
) -> Callable:
    """Helper for :meth:`.Computer.cache`.

    Parameters
    ----------
    computer : Computer, optional
        If supplied, the ``config`` dictionary stored in the Computer is used to look
        up values for `cache_path` and `cache_skip`, at the moment when `func` is
        called.
    cache_path : os.Pathlike, optional
        Directory in which to store cache files.
    cache_skip : bool, optional
        If :obj:`True`, ignore existing cache entries and overwrite them with new
        values from `func`.

    See also
    --------
    hash_args
    hash_code
    """
    log.debug(f"Wrapping {func.__name__} in Computer.cache()")

    # Wrap the call to load_func
    def cached_load(*args, **kwargs):
        try:
            # Retrieve cache settings from the `computer`
            # Only do this at time of execution, to allow the cache_path to be adjusted
            config = unquote(computer.graph["config"])
        except AttributeError:
            # No `computer` provided; use values from arguments
            config = dict()

        _dir = config.get("cache_path", cache_path)
        _skip = config.get("cache_skip", cache_skip)

        if not _dir:
            _dir = Path.cwd()
            log.warning(f"'cache_path' configuration not set; using {_dir}")

        # Parts of the file name: function name, hash of arguments and code
        name_parts = [func.__name__, hash_args(*args, hash_code(func), **kwargs)]
        # Path to the cache file
        path = _dir.joinpath("-".join(name_parts)).with_suffix(".pkl")

        # Shorter name for logging
        short_name = f"{name_parts[0]}(<{name_parts[1][:8]}…>)"

        if not _skip and path.exists():
            log.info(f"Cache hit for {short_name}")
            with open(path, "rb") as f:
                return pickle.load(f)
        else:
            log.info(f"Cache miss for {short_name}")
            data = func(*args, **kwargs)

            with open(path, "wb") as f:
                pickle.dump(data, f)

            return data

    # Update the wrapped function with the docstring etc. of the original
    update_wrapper(cached_load, func)

    return cached_load



def test_import_pyam():
    """.compat.pyam.computations is populated only if pyam itself is installed.

    Unlike the tests in :mod:`.test_pyam`, this test should pass regardless of whether
    or not pyam is installed.
    """
    from genno.compat.pyam import HAS_PYAM, computations

    # Same value, either True or False
    assert HAS_PYAM is hasattr(computations, "as_pyam")



from genno import configure
from genno.util import REPLACE_UNITS


def test_configure_units(caplog):
    # Warning is logged on invalid definitions
    configure(units=dict(define="0 = [0] * %"))
    assert 1 == len(caplog.records) and "WARNING" == caplog.records[0].levelname

    # Unit replacements are stored
    configure(units=dict(replace={"foo": "bar"}))
    assert REPLACE_UNITS.pop("foo") == "bar"
